#ifndef _JSCRIPT_H_
#define _JSCRIPT_H_

#include <algorithm>
using namespace std;

#include "DLConnectedComponents.h"
#include "DLImage.h"      // includes for Doclib  
#include "DLComponent.h"

#include "JMorphology.h"
#include "JStatistics.h"
#include "JHamming.h"

#define RESCALE_HEIGHT 30
#define RESCALE_WIDTH 30
#define MAX_FILENAME_LEN 500
#define BITWISE_HALFWAY_POINT 25 /* based on 20x20 scaling */

const static  char DL_JARGON_VERSION[] =  "JARGON_BETA_1.0 - $Name:  $";

/*! \mainpage JARGON
  <br> Overview
  <br>------------

  The JargonDOCLIB add-on implements a machine text script identification
  algorithm.  This determination is done directly from the image, prior to OCR.
  Knowledge of the script of an image can be useful in filtering and routing of text
  image to followon processing.  JARGON also attempts to classified the orientation of
  the image, although because this classification is highly biased to the upright direction,
  it may be better to use a different technique.  

  While the JargonDOCLIB add-on contains all of the 
  software needed to retrain the algorithm on desired scripts, this package includes 
  two pretrained set of data files.  One of these allows for script identification amongst
  a set of 13 scripts (Amharic, Arabic, Armenian, Burmese, Chinese, Cyrillic, Devanagari, 
  Greek, Hebrew, Japanese, Korean, Latin, and Thai).  
  The other will discriminate between Latin and 
  Arabic scripts in all four directions.  In this case the JARGON script orientation classification
  capability is not used, rather, the non-upright Latin and Arabic scripts are treated as separate
  scripts during training.

  JARGON uses a template matchingn technique to discriminate between scripts.  JARGON also includes an 
  "unknown" classification.  The template matching technique serving as a basis for JARGON was initially developed at Los Alamos
  National Laboratory (see Automatic Script Identification from Images using Cluster-based
  Templates, Proc., 3rd Int'l Conf. on Document Analysis and Recognition, pp 378-381, Montreal, Canada, August 14-16, 1995).  

  There are three major components (programs) currently used to train and then run script identification.
  Initially, the JargonTrainingDriver program is used to cluster connected components occuring within a
  training set of images.  The connected components from each training image for a script are extracted and are rescaled to a common size,
  currently 20x20.  Each comoponent is then compared to the clusters already formed.  If it is significantly different
  from the existing clusters, it becomes the basis for a new cluster.  Otherwise it is added to and existing cluster.  
  The metric used to measuer similarity is the hamming distance.  Currently, for 20x20 sized components, a hamming
  distance threshold of 111 is used.  After the clustering process is completed, the clusters( which are greyscale 
  because of averaging) ar thresholded at 50% into binary values.  This process is repeated for each script.

  The second part of training, performed by the JargonReliabilityDriver program, 
  consists of assigning a metric to each cluster in each of the scripts.  It is 
  referred to as the "reliability" number for the script.  This reliability number serves as a measure of the 
  uniqueness of a symbol to a script.  It is computed by, for each connected component in all the training images
  for all the scripts, finding the best match cluster.  The reliability number for a cluster is then defined
  as the ratio of the number of symbols best matching the cluster that are of that clusters script  divided by the
  total number of symbols best matching the cluster (   matches to cluster that are of clusters script/total matches)

  Once the first two programs have been run to generate the data files necessary to perform script identification, 
  the JargonScriptIDDriver can be used to perform script identification for an unknown image.  This program 
  randomly selects a set of connected components from an image.  The components are filtered based on density, location
  on the page, and size.  For each of the remaining symbols, the best match from amongst all the clusters for 
  each script are computed.  During this process a weighted sum is computed for each of the scripts.  For
  each best fit computed, the sum for the corresponding script is incremented by the reliability number corresponding
  to the best fit cluster.  Once all of the selected symbols from the image being classified have been computed
  the distribution of the scores for each script is used to determing the image script classification.  In
  general the script with the highest score is chosen.  If, however, no one script scores strongly (the standard
  deviation across scripts is less then 1.0) the image may be classified as unknown. At this point 200 symbols are being used 
  to classify an images.  This is a configurable parameter.  To select an orientation for an image, JARGO
  N computes the script scores for all four directions for the first 50 components.  If one direction is showing a substantially 
  higher standard deviation, indicating tha tone or two scripts are clearly beginning to emerge as predominant,
  that direction is selected.  Otherwise JARGON assumes the image is upright.



  <br> Input Image Format
  <br> -------------------
 
  This algorithm operates in the binary domain, however, the software
  will automatically use the DOCLIB functions 
  DLBitsPerPixelConverter::DLBitsPerPixelConverter::dlConvertImage to 
  convert color or grayscale images
  to binary.  At this point there is no way to specify or select an a
  lternate color/grayscale binary conversion.  If another method is desired the 
  images should be converted prior to providing them to the DLLogoTokenMatchDOCLIB 
  functions.
*/


/** Structure used to store all information pertaining to a particular script.  Most
 * of the fields are used to store the results of particular training and reliability 
 * computations 
 **/
/// Structure used to store all information pertaining to a particular script.
typedef struct _language_results {
   ///script name array.  max 50 languages
   char language[50];
   /// total symbol number
   int total_symbol_count;
   /// number of symbols correctly matched to this script
   int correctly_matched;   /*  */
   /// number of symbols incorrectly  matched to this script
   int incorrectly_matched; /*  */
   /// number of symbols incorrectly matched to other scripts
   int matched_to_other;    /*  */
   /// previous language
   char *prev;
   /// next language
   char *next;
   /// spare
   char spare[50];
} language_results;

/**
 * The JScript class contains all the methods needed to customize the JARGON script 
 * identification algorithm and to run the algorithm
 */

/// Class used to customize and run script identification

class JScript {

   public:
   /**
    *  read_script_data() function populates a lanugate profile data structure from
    *  trained language scripts.
    *
    * \param language_list_path text file contains a list of trained languages
    * \param language_array a pointer to an empty language_profile pointer.  
    *                read_script_data will populate this data structure from the language file.
    * \param language_count the number of languages in the language_list_path file.  This will be populated.
    * \param scaled_height desire height for  components
    * \param scaled_width desire width for  components
    */
   void read_script_data(char *language_list_path, language_profile **language_array, int *language_count, int scaled_height, int scaled_width);
   /**
    *  read_script_data() function populates a lanugate profile data structure from
    *  trained language scripts.
    *
    * \param language_list_path text file contains a list of trained languages
    * \param language_array a pointer to an empty language_profile pointer.  
    *                read_script_data will populate this data structure from the language file.
    * \param language_count the number of languages in the language_list_path file.  This will be populated.
    * \param scaled_height desire height for  components
    * \param scaled_width desire width for  components
    * \param threshold
    * \param gray_values
    */	
   static void ReadScriptData(char *language_list_path, language_profile **language_array, int *language_count, int scaled_height, int scaled_width, float *threshold,
                              int **gray_values);

   /** 
    * jarGetScriptNo returns the predefined ID for each of the scripts
    * \param script name of the script
    * \return int <pre>
    *    0  amharic
    *    1  arabic
    *    2 armenian
    *    3 burmese
    *    4 chinese
    *    5 cyrillic
    *    6 devanagari
    *    7 greek
    *    8 hebrew
    *    9 japanese
    *    10 korean
    *    11 latin
    *    12 thai
    *    13 nodata
    *    14 others </pre>
    */
   int jarGetScriptNo(char *script);

   /**
    * constructor
    * \param script_data_file_path path to the pre-trained script files
    * 
    **/
   JScript(char *script_data_file_path);  // for class instantiation
   /**
    * Destructor
    **/
   ~JScript();                            // destructor
   /**
    * Assign a new border size for documents.  Connected components within
    * the border will not be used in classifying the script.
    * \param top top margin
    * \param left left of margin
    * \param bottom bottom margin
    * \param right right margin
    **/ 
   void  changeBorders( int top, int left, int bottom, int right);

   /**  clearDistribution sets the scores for all scripts to zero
    */

   void  clearDistribution();   
   /**
    *  disableScript allows for selected scripts to be turned off.  This
    * option should be used judiciously.  Disabling too many scripts will
    * result in the performance of script identification to degrade.  This
    * function is most useful to turn off one of two similar scripts.  For
    * example if Japanese and Chinese are both part of the scripts JARGON
    * is trained on turning of one may improve accuracy.
    **/
   void  disableScript(int script_no);  
   /** 
    * Allows for enabling of a script that was previously disabled using
    * the disableScript method.
    **/
   void  enableScript(int script_no);   
   /** 
    * Enables all the scripts
    **/
   void  enableAllScripts();            
   /**
    * Allows for specification of the page orientaions JARGON will consider.
    * the newdirections array contains a 0/1 flag indicating if the directions
    * (0,90,180,27) are enabled.
    **/
   void  setDirections(int newdirections[4]);
   /** Specifies the number of symbols to be used for script identification
    */
   void  setSymbol_count(int symbol_limit);  
   /** Allows for modification of the threshold used in determining if a non-
    * upright orientation is selected 
    */
   void  setOrientThreshold(int thresh);     
   /**
    * GetScript ID will determine the type of script in a giving image.
    * \param testImageP an image printer to the image to be identified. The input image must be a binary(Black/White) image.
    *    DOCLIB will throw an exception when image is in grayscale or color.
    * \return a string of the script identified
    */
   char *getScript(DLImage &testImageP);     

   /**
    * return the directions array
    * \return pointer to directions[]
    */
   int   *getDirections()      { return (&directions[0]); }
   /** returns the symbol distribution (float) containing the scores
    * for the scripts
    */
   float *getDistribution()    { return (&symbol_distribution[0]); }
   /** Returns the highest scoring script.  Note, this may or may not
    * be the JARGON classification for the page.  If another script scored
    * high as well JARGON may classify the image as unknown
    */

   char  *getFirst()           { return (&best_fit_script[0]); }
   /** Returns the second highest scoring script.  
    */
   char  *getSecond()          { return (&second_best_script[0]); }
   /** Returns JARGON's orientation classification
    */
   int    getOrientation()     { return (_orientation); } 
   char  *getOrient_char()     { return (&orient_char[0]); }
   int   *getProb()            { return (probability); }
   /** Returns the number of scripts that JARGON is trained on
    */
   int    getScriptCount()     { return (language_count); } 

   /** Returns the number of symbols that were used to perform 
    * script identification
    */
   int    getSymbol_count()    { return (symbol_count); }

   /** Returns the number of symbols that were used to perform 
    * script identification
    */
   int	   getSymbolsUsed()	    { return (symbols_used_for_id_count); }
   /** Returns the threshold used in determining if a non-upright
    * orienation is presentw
    */
   int    getOrientThreshold() { return (orient_threshold); }
   
   static string dlGetVersionInfo();

   private:

   //   
   // symbol_distribition stores the scores for hte scripts

   float    *symbol_distribution;   

   // random number array used to randomly select components from a page
   float    *randoms;            

   // Lookup table used to speed up calculation of hamming distance
   int     *hamming_lookup;       

   int      *probability;           

   /** Data structute used to store data used in script identification.
    * It is initialized by ReadScriptData
    */

   language_profile  *first_lang;  
   int       language_count;
   int       min_comp_size;
   int       max_height;
   int       max_width;
   int       min_height;
   int       min_width;
   int       symbol_count;
   int		  symbols_used_for_id_count;
   int       min_symbol_count;
   float     blob_percent;
   int       rescale_height;
   int       rescale_width;
   int       upper_offset;
   int       lower_offset;
   int       left_offset;
   int       right_offset;
   float     reliability_threshold;
   float     template_threshold;
   float     separation_threshold;
   float     mlp_thresh;
   int       orient_threshold;   // e.g.  50 or 100 or 150
   int       hamming;
   int       directions[4];	// 0,90,180,270 deg 
   //    class instantiation sets up / initializes fields 
   //    symbol_distribution through directions.

   //    calling getScript establishes result values for fields 
   //    _orientation through second_best_script.
   int       _orientation;
   char      orient_char[2];
   char      script_id[100];        // getScript() value
   char      best_fit_script[100];
   char      second_best_script[100];

   char      script_selection_algorithm[100];
};


#endif //_JSCRIPT_H_
