// layout.cpp: implementation of the Clayout class.
//
//////////////////////////////////////////////////////////////////////

#include "layout.h"
#include "../PageLayoutLib/objectInfoClass.h"
#include "../Utility_hm/Kmeans.h"
#include "../Utility_hm/TMatrix.h"
#include "../ObjectVector_hm/objcluster.h"
#include <math.h>

#ifndef M_PI
#define M_PI 3.1415926536
#endif

/**
 * Constructor
 **/
Clayout::Clayout()
{
   m_pObjectList = NULL;
   m_pClusterList = NULL;
	
   m_nbClusters = 0;
   m_nbObjects  = 0;

}

/**
 * Destructor
 **/
Clayout::~Clayout()
{
   if(m_pObjectList)	delete []m_pObjectList;
   if(m_pClusterList)	delete []m_pClusterList;

}

/**
 * ImportObjects constructs object list from input data structure
 * \param obj the input data structure holding the information to be imported
 * \param objcount the number of objects to be imported
 **/
void Clayout::ImportObjects(struct objectInfo *obj, int objcount)
{
   Reset();

   // make the majority textlines horizontal
   m_nbObjects = objcount;

//	if(m_pObjectList)	delete m_pObjectList;
   m_pObjectList = new objvector[m_nbObjects];

   // import long data to double data matrix
   CTMatrix<double> locations(m_nbObjects,4);

//	m_pageMinfontheight = 1000;
   int i;
   for (i=0;i<m_nbObjects;i++)
   {		
      m_pObjectList[i].Type = obj[i].type;

      m_pObjectList[i].FontHeight = (double)obj[i].int_qualifier;	
		
      locations[i][0] = obj[i].location[0].x;
      locations[i][1] = obj[i].location[0].y;
      locations[i][2] = obj[i].location[1].x;
      locations[i][3] = obj[i].location[1].y;
		
      m_pObjectList[i].Length =	sqrt( (double)(obj[i].location[0].x - obj[i].location[1].x)*(obj[i].location[0].x - obj[i].location[1].x)
                                      +(obj[i].location[0].y - obj[i].location[1].y)*(obj[i].location[0].y - obj[i].location[1].y));
		
      m_pObjectList[i].Angle  =   obj[i].location[0].x - obj[i].location[1].x==0 ? - M_PI/2 : 
         atan((double)(obj[i].location[0].y - obj[i].location[1].y)/(double)(obj[i].location[0].x - obj[i].location[1].x));
		
//		m_pageMinfontheight = min(m_pageMinfontheight, m_pObjectList[i].FontHeight);
   }

   Rotate(locations);	

   // find the bounding rectangle
   // now all objects have been rotated
   double minx = 1000, miny = 1000;
   double maxx = 0,    maxy = 0;
   for (i=0;i<m_nbObjects;i++)
   {
      m_pObjectList[i].Centroid_X =	(locations[i][0] + locations[i][2])/2;
      m_pObjectList[i].Centroid_Y =	(locations[i][1] + locations[i][3])/2;

      // record min/max x and min/max y for normalization
      double tmp_minx = min(locations[i][0], locations[i][2]);
      minx = min(minx, tmp_minx);

      double tmp_miny = min(locations[i][1], locations[i][3]);
      miny = min(miny, tmp_miny);

      double tmp_maxx = max(locations[i][0], locations[i][2]);
      maxx = max(maxx, tmp_maxx);

      double tmp_maxy = min(locations[i][1], locations[i][3]);
      maxy = max(maxy, tmp_maxy);		

   }

   m_pageheight = maxy - miny;
   m_pagewidth  = maxx - minx;
	
   // before scaling, make the bounding rectangle square
//	double side = max(m_pageheight,m_pagewidth);
   double side = 3000;

   double refer_x = minx;
   double refer_y = miny;
   for (i=0;i<m_nbObjects;i++)
   {
      // now the length is scaled to a unit bounding square 
      m_pObjectList[i].Length /= side;

      // put the page lefttop point to origin, and scale, so x and y belong to[0,1]
      m_pObjectList[i].Centroid_X = (m_pObjectList[i].Centroid_X - refer_x)/side;
      m_pObjectList[i].Centroid_Y = (m_pObjectList[i].Centroid_Y - refer_y)/side;	
		
      // also scale the font height
      m_pObjectList[i].FontHeight /= side;

/*		// all angles are normalized against x_axis
		int sign = m_pObjectList[i].Angle >=0 ? 1 : -1;
		if(sign * m_pObjectList[i].Angle > PI/4)
                m_pObjectList[i].Angle = (PI/2 - sign * m_pObjectList[i].Angle) * sign;
*/
      m_pObjectList[i].Angle = m_pObjectList[i].Angle/M_PI + 0.5;  

   }

}

/**
 * SaveClustersToDisk dump information to disk
 * \param fname the filename where the information is to be write to
 * \param saveweights flag indicating whether write the weiths member into file or
 * use default value.
 **/
void Clayout::SaveClustersToDisk(string fname, bool saveweights )
{
   ofstream	f;
   f.open( fname.c_str(),ios::out );
   if( f.rdstate() & ios::failbit ) 
   {
      printf("open file error!");
      f.close();
      return;
   }

   f << "nbClusters: " <<m_nbClusters << endl;

   //write the vector
   int i;
   for(i=0; i<m_nbClusters; i++)
      m_pClusterList[i].write_stream(f, saveweights);	

   f.close();
}

/**
 * LoadClustersFromDisk loads cluster information from file
 * \param fname the filename where cluster information will be loaded
 **/
void Clayout::LoadClustersFromDisk(string fname)
{
   Reset();

   ifstream	f;
   f.open( fname.c_str(),ios::in );
   if( f.rdstate() & ios::failbit ) 
   {
      printf("open file error!");
      f.close();
      return;
   }

   char buf[80];
   f >>buf; f>> m_nbClusters ;

   m_pClusterList = new objcluster[m_nbClusters];

   //read the vector
   int i;
   for( i=0; i<m_nbClusters; i++)
      m_pClusterList[i].read_stream(f);	

   f.close();
}


/**
 * Rotate rotates a document so that most of its text line objects are horizontal 
 * \param obj the text line objects to be rotated
 **/
void Clayout::Rotate(CTMatrix<double> &obj)
{
#define BINSIZE 2.5   //in degree

   //angle histogram : each bin contains 2.5 degree = 2.5/180*PI arcangle
   int nbin = int(180.0/BINSIZE);
   double interval = M_PI/nbin;
	
   int * histo = new int [nbin] ;
   int i;
   for( i=0;i<nbin; histo[i++] = 0);

   for(i=0;i<m_nbObjects; i++)
   {
      //Angle is between [-PI/2, PI/2)			
      int index = int((m_pObjectList[i].Angle + M_PI/2)/interval+0.5);

      histo[index]++;
   }

   //find the dominant neighboring two bins 
   int max = 0, index = -1;
   for(i = 0;i<71;i++)
      if(histo[i]+histo[i+1] >max)
      {
         max = histo[i]+histo[i+1];
         index = i;
      }

   double dom_angle = ((index*histo[index]+(index+1)*histo[index+1])/(double)max)*interval- M_PI/2;

   delete histo;

   // do not rotate if angle is in [-2.5, +2.5]
   if(fabs(dom_angle) < interval*2) 
      dom_angle = 0;	

   // rotate all objects by angle = dom_angle
   for(i=0;i<m_nbObjects; i++)
   {
      m_pObjectList[i].Angle -= dom_angle;
		
      //rotate all objects
      double newx = cos(dom_angle)*obj[i][0] - sin(dom_angle)*obj[i][1];
      double newy = sin(dom_angle)*obj[i][0] + cos(dom_angle)*obj[i][1];

      obj[i][0] = newx;
      obj[i][1] = newy;

      newx = cos(dom_angle)*obj[i][2] - sin(dom_angle)*obj[i][3];
      newy = sin(dom_angle)*obj[i][2] + cos(dom_angle)*obj[i][3];

      obj[i][2] = newx;
      obj[i][3] = newy;
   }

}

/**
 * Reset release memory and reset all working variable to default values
 **/
void Clayout::Reset()
{
   if(m_pObjectList)	delete []m_pObjectList;
   if(m_pClusterList)	delete []m_pClusterList;

   m_pObjectList = NULL;
   m_pClusterList = NULL;

   m_nbClusters = 0;
   m_nbObjects  = 0;

}

/**
 * operator = overides assign operator
 **/
Clayout & Clayout::operator =(Clayout &a)
{
   Reset();

   m_index = a.m_index;
   m_nbClusters = a.m_nbClusters;
   m_nbObjects = a.m_nbObjects;
	

   m_pageheight = a.m_pageheight;
   m_pagewidth = a.m_pagewidth;
	

   if(m_nbClusters > 0)
   {
      m_pClusterList = new objcluster[m_nbClusters];
      memcpy(m_pClusterList, a.m_pClusterList, sizeof(objcluster)*m_nbClusters);
   }

   if(m_nbObjects >0)
   {
      m_pObjectList = new objvector[m_nbObjects];
      memcpy(m_pObjectList, a.m_pObjectList, sizeof(objvector)*m_nbObjects);
   }

   // m_pPairList = a.m_pPairList;
   /*m_avg = a.m_avg;
     m_sigma = a.m_sigma;
     m_weight = a.m_weight;*/

   return *this;
}

/**
 * BoundingBox finds the minimum bounding rectangle of a set of text line objects
 * \param obj the object set
 * \param objcount the number of objects in the set
 * \param r the bounding rectanle to be returned
 **/
void Clayout::BoundingBox(objectInfo *obj, int objcount, rectangle &r)  
{

   Reset();

   // make the majority textlines horizontal
   m_nbObjects = objcount;

//	if(m_pObjectList)	delete m_pObjectList;
   m_pObjectList = new objvector[m_nbObjects];

   // import long data to double data matrix
   CTMatrix<double> locations(m_nbObjects,4);

   //m_pageMinfontheight = 1000;
   int i;
   for ( i=0;i<m_nbObjects;i++)
   {		
      m_pObjectList[i].Type = obj[i].type;

      m_pObjectList[i].FontHeight = (double)obj[i].int_qualifier;	
		
      locations[i][0] = obj[i].location[0].x;
      locations[i][1] = obj[i].location[0].y;
      locations[i][2] = obj[i].location[1].x;
      locations[i][3] = obj[i].location[1].y;
		
      m_pObjectList[i].Length =	 sqrt((double)(obj[i].location[0].x - obj[i].location[1].x)*(obj[i].location[0].x - obj[i].location[1].x)
                                      +(obj[i].location[0].y - obj[i].location[1].y)*(obj[i].location[0].y - obj[i].location[1].y));
		
      m_pObjectList[i].Angle  =   obj[i].location[0].x - obj[i].location[1].x==0 ? - M_PI/2 : 
         atan((double)(obj[i].location[0].y - obj[i].location[1].y)/(double)(obj[i].location[0].x - obj[i].location[1].x));
		
      // m_pageMinfontheight = min(m_pageMinfontheight, m_pObjectList[i].FontHeight);
   }

   Rotate(locations);	

   // find the bounding rectangle
   // now all objects have been rotated
   double minx = 4000, miny = 4000;
   double maxx = 0,    maxy = 0;
   for (i=0;i<m_nbObjects;i++)
   {
      // record min/max x and min/max y for normalization
      double tmp_minx = min(locations[i][0], locations[i][2]);
      minx = min(minx, tmp_minx);

      double tmp_miny = min(locations[i][1], locations[i][3]);
      miny = min(miny, tmp_miny);

      double tmp_maxx = max(locations[i][0], locations[i][2]);
      maxx = max(maxx, tmp_maxx);

      double tmp_maxy = min(locations[i][1], locations[i][3]);
      maxy = max(maxy, tmp_maxy);		

   }

   r.top = miny;
   r.left = minx;
   r.bottom = maxy;
   r.right = maxy;
}


/**
 * KmeansObjClustering uses kmeans clustering to cluster textline vectors
 * \param pLayouts the layout set whose text lines are to clustered
 * \param nObj the total number of text lines in the set
 * \param nDoc the number of document pages involved in the set
 * \param nCluster the targeted number of clusters
 * \param maxiteration the maximum iterations in kmeans execuation
 * \param sigma the default standard deviation for a cluster, used when the computed standard
 * deviation is zero.
 * \param use_zero_flag a boolean flag, if true, the document containing no member of a cluster will 
 * be counted in computing the mean and variance of member occurence for average document
 **/
void Clayout::KmeansObjClustering(Clayout * pLayouts, int nObj, int nDoc, int nCluster, 
                                  int maxiteration, objvector sigma, bool use_zero_flag )
{
   int nDim = 5;

   CKmeans kmeans(nObj,nDim,nDim,false,nCluster); // target #clusters 
		
   CTArray<int> iDoc(nObj); // record document id for each object
		
   // dump all objects into CKmeans datastructure
   int nCount;
   int j;
   for( nCount=0,j=0;j<nDoc;j++)
   {
      int j_nbObj = pLayouts[j].m_nbObjects;
      int k;
      for( k=0;k<j_nbObj;k++)
      {
         kmeans.Pattern[nCount][0] = pLayouts[j].m_pObjectList[k].Angle ;
         kmeans.Pattern[nCount][1] = pLayouts[j].m_pObjectList[k].Centroid_X ;
         kmeans.Pattern[nCount][2] = pLayouts[j].m_pObjectList[k].Centroid_Y ;
         kmeans.Pattern[nCount][3] = pLayouts[j].m_pObjectList[k].FontHeight ;
         kmeans.Pattern[nCount][4] = pLayouts[j].m_pObjectList[k].Length ;
//				kmeans.Pattern[nCount][5] = pLayouts[j].m_pObjectList[k].Type;

         iDoc[nCount] = j;

         nCount++;
      }
   }		
		
//		kmeans.SetConstraint(5, 1);		// object type must be agree
   kmeans.SetConstraint(0, 0.1*(sigma.Angle));	// Angle should be similar
   kmeans.SetConstraint(3, 0.005*(sigma.FontHeight));	//FontHeight difference should be small
   CTArray<double> tmp(nDim);		
   tmp[0] = sigma.Angle/10;
   tmp[1] = sigma.Centroid_X/10;
   tmp[2] = sigma.Centroid_Y/10;
   tmp[3] = sigma.FontHeight/10;
   tmp[4] = sigma.Length/10;
   kmeans.SetDefaultStdDev(tmp);
   kmeans.RunKMeans(maxiteration);

   // output clustering results into pClusters
   m_pClusterList = new objcluster[nCluster];
   int s;
   int ss;
   for( s=0, ss=0;s<nCluster;s++)
   {
      if(kmeans.Cluster[s].NumMembers/(double)nDoc <0.3 ) continue;			

      m_pClusterList[ss].m_centroid.Angle			= kmeans.Cluster[s].Center[0];
      m_pClusterList[ss].m_centroid.Centroid_X	= kmeans.Cluster[s].Center[1];
      m_pClusterList[ss].m_centroid.Centroid_Y	= kmeans.Cluster[s].Center[2];
      m_pClusterList[ss].m_centroid.FontHeight	= kmeans.Cluster[s].Center[3];
      m_pClusterList[ss].m_centroid.Length		= kmeans.Cluster[s].Center[4];
//			m_pClusterList[ss].m_centroid.Type		    = kmeans.Cluster[s].Center[5];

      m_pClusterList[ss].m_sigma.Angle		= kmeans.Cluster[s].SDev[0];
      m_pClusterList[ss].m_sigma.Centroid_X	= kmeans.Cluster[s].SDev[1];
      m_pClusterList[ss].m_sigma.Centroid_Y	= kmeans.Cluster[s].SDev[2];
      m_pClusterList[ss].m_sigma.FontHeight	= kmeans.Cluster[s].SDev[3];
      m_pClusterList[ss].m_sigma.Length		= kmeans.Cluster[s].SDev[4];

      m_pClusterList[ss].m_nbObjects			= kmeans.Cluster[s].NumMembers;

      // build the document-object histogram for Cluster s
      CTArray<int> histo(nDoc);
      int i;
      for( i=0;i<m_pClusterList[ss].m_nbObjects;i++)
      {
         int obj_index = kmeans.Cluster[s].Member[i];
         int doc_index = iDoc[obj_index];
         histo[doc_index] ++;
      }

      //count the document that does not include any of the members
      int j;
      int nZero;					//nZero stnads for non-Zero
      for( j=0, nZero = 0;j<nDoc; j++)
         if(histo[j] > 0) nZero ++;

      if(!use_zero_flag) nZero = nDoc;

      //compute the mean and variance of the #obj per doc
      m_pClusterList[ss].m_avgnbObjperdoc = m_pClusterList[ss].m_nbObjects/(double)nZero;
      m_pClusterList[ss].m_sigmaObjperdoc = 0;
      int k;
      for( k=0;k<nDoc;k++) 
      {
         if(use_zero_flag && histo[k]==0) continue;
         m_pClusterList[ss].m_sigmaObjperdoc += (histo[k] - m_pClusterList[ss].m_avgnbObjperdoc)
            *(histo[k] - m_pClusterList[ss].m_avgnbObjperdoc);
      }

      m_pClusterList[ss].m_sigmaObjperdoc /= (double)nZero;
      if(m_pClusterList[ss].m_sigmaObjperdoc <1e-4)
         m_pClusterList[ss].m_sigmaObjperdoc = 0.0001;
		
      ss ++;
   }

   m_nbClusters = ss;		

}


/**
 * Weight_Clusters_against scores/weights wanted clusters against unwanted clusters (referring clusters):
 * find unwanted clusters within fixed range of each wanted cluster;
 * use the ratio #avg_obj_per_doc_for_a_wanted_cluster/(#avg_obj_per_doc_within_this_range)
 * \param pLayout the layout set containing the referring clusters
 * \param probability_range the threshold used for determing whether an unwanted cluster is a 
 * neighbor of a wanted cluster
 **/
void Clayout::Weight_Clusters_against(Clayout * pLayout, double probability_range)
{
   objcluster * pWanted = m_pClusterList;
   objcluster * pUnWanted = pLayout->m_pClusterList;

   int nWanted = m_nbClusters;
   int nUnWanted = pLayout->m_nbClusters;
   int i;
   for( i=0;i<nWanted;i++)
   {
      if(pWanted[i].m_avgnbObjperdoc < 0.0001)
      {
         pWanted[i].m_weight = 0;
         continue;
      }

      double sum_dist = 0;
      double nNeighbor = 0;  //weighted sum of number of neighboring unwanted members
      for (int j=0;j<nUnWanted;j++)
      {
         double probability = pWanted[i].Probability(pUnWanted[j].m_centroid);

         if(probability > probability_range)
            nNeighbor += pUnWanted[j].m_avgnbObjperdoc;
      }

      pWanted[i].m_weight = pWanted[i].m_avgnbObjperdoc /(pWanted[i].m_avgnbObjperdoc+ nNeighbor);
			
      /*	objvector diff = pWanted[i].m_centroid - pUnWanted[j].m_centroid;
                double dist = diff.norm_2(0.3, 0.3 ,0.15, 0.15,0.1);
			
                if(dist < FIXED_RANGE)
                { 
                if(weighted_bydist)
                {
                sum_dist += dist;
                nNeighbor += pUnWanted[j].m_avgnbObjperdoc* dist;
                }
                else
                nNeighbor += pUnWanted[j].m_avgnbObjperdoc;
                }
		}
			
		if(sum_dist ==0) sum_dist = 1;
		double ratio = nNeighbor/sum_dist;		

		pWanted[i].m_weight = pWanted[i].m_avgnbObjperdoc/(pWanted[i].m_avgnbObjperdoc+ratio);
      */

   }
	
}


/**
 * NormalizeObjects normalizes all object vectors from all documents by simple division
 * \param pLayouts the layout set containing the objects
 * \param nDoc the number of documents in the layout set
 * \param avg the mean of the all the objects in the layout set
 * \param sigma the variance of the objects
 **/
void Clayout::NormalizeObjects(Clayout * pLayouts, int nDoc, objvector &avg, objvector &sigma)
{
   int nbObjects =0 ;
   int i= 0;
   for( ; i<nDoc;i++)
   {
      int j;
      for( j=0;j<pLayouts[i].m_nbObjects;j++)
      {
         avg = avg + pLayouts[i].m_pObjectList[j];
         nbObjects ++;
      }
   }
	
   avg = avg / (double) nbObjects;

	
   int j;
   for(i=0;i<nDoc;i++)
      for( j=0;j<pLayouts[i].m_nbObjects;j++)
      {
         objvector diff = pLayouts[i].m_pObjectList[j]- avg;

/*			// deal with angle differently
			double sign = diff.Angle <0 ? -1 : 1;
			double abs_diff = fabs(diff.Angle);
			diff.Angle = abs_diff >0.5? (PI - abs_diff)*sign : diff.Angle;
*/
         sigma = sigma + diff.dotabs();
      }	
	
   sigma = sigma / (double)nbObjects;

/*	for(i=0;i<nDoc;i++)
        for(int j=0;j<pLayouts[i].m_nbObjects;j++)
        {
        objvector diff = pLayouts[i].m_pObjectList[j]- avg;

        // deal with angle differently
        double sign = diff.Angle <0 ? -1 : 1;
        double abs_diff = fabs(diff.Angle);
        diff.Angle = abs_diff >0.5? (PI - abs_diff)*sign : diff.Angle;

        pLayouts[i].m_pObjectList[j] = diff.dotdivide(sigma);		
		
        }
*/

}
