#ifndef __objectInfoClass_H
#define __objectInfoClass_H


#ifndef _JSCRIPT_H_
#include <JScript.h>
#endif

#ifndef __getLines_H
#include <getLines.h>
#endif

#ifndef __polyClass_H
#include <polyClass.h>
#endif


/** The objectType enumeration represents each of the possible object type used
 * by the page layout software.  Not all of the object types enumerated are currently
 * being used 
 **/
enum objectType {nullObject = 1, taggedForDelete, textLineCount, script, caereWord,  pageType, textLine,
                 scansoft_flowed_text,scansoft_table,scansoft_graphics};


/** The ObjectInfo data structure is used to capture generic information
 * about an object or feature calculated for a document page.  The string
 *	qualifier and int_qualifier will be specific to the type of object
 *	and may not always be used. pointLocation are the two points used
 *	to represent the object spacially.  The values may or may not
 *	be specific to type of object.  For text lines it may be endpoints.
 *	Other possibilities in include 1st order moments. Essentially each
 *	object is represented in space by the lined formed by joining
 *	the two points 
 */
struct objectInfo
{
   objectType type;
   char *qualifier;   /* generic string qualifier, specific to object type/ */
   int int_qualifier; /* generic string qualifier, specific to object type */
   float upper_ext;   /* only used for text lines */
   float lower_ext;   /* only used for text lines */
   struct pointLocation location[2]; /* Location of two points used to represent an object */
//CPoint location[2];

   unsigned int occuranceCount; /* used occasionally to count occurances */
};
/**
 * The objectInfoClass is used to contain a collection of information/features calculated
 * for a document image.
 */
class objectInfoClass {

   public:
/** This method creates and instance of an objectInfoClass.  The following paramenters are used:
 * filename - path the the image file
 * scriptInfo - an instance of the JScript classed that is used to calculate page level script
 *             identification
 * line_count_limit - maximum number of text lines to include in the class
 * output_directory - directory where output data files containing the object info. are stored.
 *             Prior to computing the needed features, the software will look in the specified
 *			 data directory to see if a .dat file corresponding to the image exists.  If so,
 *			 the data in that file will be read in instead of recalculating the features. 
 * pixelDataFlag - indicates whether the pixel data for the image should be stored in the class
 * scansoftFlag - unused at this point.  Can be used, with code modifications to indicate if 
 *			scansoft bounding boxes should be included.
 * dctFlag - indicates if features relating to interline spacing should be included.
 *
 */

   objectInfoClass::objectInfoClass(std::string filename,
                                    JScript *scriptInfo,int line_count_limit, char *output_directory,
                                    int pixelDataFlag,int scansoftFlag,int dctFlag);

/** This method creates and instance of an objectInfoClass.  The following paramenters are used:
 * filename - path the the image file
 * scriptInfo - an instance of the JScript classed that is used to calculate page level script
 *             identification
 * scansoftFlag - unused at this point.  Can be used, with code modifications to indicate if 
 *			scansoft bounding boxes should be included.
 * dctFlag - indicates if features relating to interline spacing should be included.
 *
 */

   objectInfoClass::objectInfoClass(std::string filename,
                                    JScript *scriptInfo,int scansoftFlag, int dctFlag);

/** This method destroys and instance of an objectInfoClass.  
 */

   objectInfoClass::~objectInfoClass();

/* getObjectCount returns the number of objects stored in the class */

   int getObjectCount() const {return( _objectCount);}

/* objectStructure returns an imageObject data structure containing all the
 *  data calculated for the class 
 */

   objectInfo *objectStructure() const{return( imageObjects);}

/* printObjectInfo prints the object information to the specified file.  If no
 *  file is specified the output is directed to stdout 
 */
   void objectInfoClass::printObjectInfo(char *filename);

/* printTextLineInfo prints the object information to the specified file.  If no
 *  file is specified the output is directed to stdout 
 */

   void objectInfoClass::printTextLineInfo(FILE *fp);

/** getImageHeight returns the height of the deskewed image 
 */

   unsigned long getImageHeight() const {return( _height);}

/** getImageWidth returns the width of the deskewed image 
 */

   unsigned long getImageWidth() const {return( _width);}

/** getImageHeight returns the height of the original image 
 */

   unsigned long getOriginalImageHeight() const {return( _originalHeight);}

/** getImageWidth returns the width of the original image 
 */

   unsigned long getOriginalImageWidth() const {return( _originalWidth);}

/** getOriginalSkew returns the skew value calculated 
 */

   double getOriginalSkew() const {return( _originalImageSkew);}

/** getPixelData returns an 8 bit/pixel representation of the pixel data 
 */

   unsigned char *getPixelData() const {return( _pixelData);}

/** getSlopeVarSqr returns the variance of the slope values squared 
 */

   double getSlopeVarSqr() const {return (_lineSlopeVarianceSqr);}

/** getLineCount returns the number of text lines calculated 
 */

   int getLineCount() const {return (_lineCount);}

/** getScript returns the image script as calculated by the JScript class 
 */

   char *getScript() {return (&_script[0]);}

/** getBestFitScript returns the best fitting image script as calculated by the JScript class.
 *   This may or may not be the same value as returned by the getScript class.  There are occasions
 *   where an image is classified as unknown.  This occurs when the best fitting script is 
 */

   char *getBestFitScript() {return (&_bestFitScript[0]);}

/** getlineDirection returns the line orientation (horizontal/vertical) of the text 
 */

   int getlineDirection() {return(_lineDirection);}

/** getscriptIdOrientation returns the image orientation as calculated by the JScript class 
 */

   int getscriptIdOrientation() {return(_scriptIdOrientation);}

/** writeG4Tiff writes out a group 4 compressed version of the original image 
 */

   void objectInfoClass::writeG4Tiff(char *fname);

/** writeG4Tiff writes out a group 4 compressed version of  image rotated  and
 *   deskewed according to the values specified by the input parameters.
 *
 */

   void objectInfoClass::writeG4TiffGray(char *fname, float *skew,
                                         int image_rotation);

/** writeG4Tiff writes out a group 4 compressed version of  image rotated  
 *   according to the values specified by the input parameters.
 *   This is not yet implemented. 
 *
 */

   void objectInfoClass::writeG4TiffGray(char *fname, int orientation);

/** getPeakValue returns the value corresponding to the first peak of the
 *   dct used to compute the interlinen spacing.  This is not yet implemented. 
 */

   float getPeakValue() const {return( _peakValue);}

/** getPeakRatio - This is not yet implemented. 
 */


   float getPeakRatio() const {return( _peakRatio);}

/** getNormLineHeightVar returns the square of the normalized height variation
 */

   double getNormLineHeightVar() const {return( _normLineHeightVar);}

/** getNormLineHeightVar returns the square of the normalized height variation
 */

   double getNormLineWidthVar() const {return( _normLineWidthVar);}

/** getNormLineIndentVar returns the square of the left indent variation
 */

   double getNormLineLeftIndentVar() const {return( _normLineLeftIndentVar);}


/** getNormLineRightIndentVar returns the square of the right indent variation
 */

   double getNormLineRightIndentVar() const {return( _normLineRightIndentVar);}

/** getAvgLineHeight returns the average line height
 */

   double getAvgLineHeight() const {return (_avgLineHeight);}


   void computeScansoftSegments(DLImage *doink,objectInfo *imageObjects ,int *index,char *filename);

   int countScansoftSegments(char *filename);

   objectInfo *computeObjectInfo(std::string fname,int pixelDataFlag);

   void computeInterlineSpacingFeatures(DLImage *doink,DLlineInfo *textLines, 
                                        int long_line_count,float *peakValue, float *peakRatio);
   void find_first_peak(double *input_array,int length,int offset, int *location);

   void computeLineStatistics( DLlineInfo *textLines, int long_line_count);

   void getSkewRotationCorrectedImageData(char *fname,float *skew, int rotation,
                                          unsigned char **ptr, int *new_height, int *new_width);

   int tbComputeDCT(double *input, int winsize, double **dct_array);

   void tbCalcMeanStddev(float *values, int value_count, double *mean, double *stddev);

   void tbStructureComputeMeanStddev(double *values, int value_count, int separation,
                                     double *mean, double *stddev);

   void tbComputeMeanStddev(double *values, int value_count, double *mean,double *stddev);



   private:

   double _lineSlopeVarianceSqr;

/** original image height */

   unsigned long _originalHeight;

/** original image width  */
   unsigned long _originalWidth;

/** image width - after deskew (if necessary) */

   unsigned long _height;

/** image height - after deskew (if necessary) */
   unsigned long _width;

   JScript *_scriptInfo;

/** pixel data from the original image - not
    deskewed or rotated **/

   unsigned char *_pixelData;
/** Horizontal Resolution **/
   float _horRes;
/** Vertical Resolution **/
   float _verRes;

/** grayscale of original pallette color image data. */

   unsigned char *_grayPixelData;
   int _objectCount;
   int _lineCount;
   int _scansoftFlag;
   /** Flag for computation of interline spacing features */
   int _dctFlag;    
   char _script[20];
   char _bestFitScript[20];
   double _originalImageSkew;
   int _lineDirection;
   int _scriptIdOrientation;
   struct objectInfo *imageObjects; 
   struct _upper_ext;
   struct _lower_ext;
   float _peakValue;
   float _peakRatio;

/** Threshold to use when performing global binarization */

   float _threshold;
/** Text line height variance divided by line height mean */
   double _normLineHeightVar;

/** Average text line height (not normalized) */
   double _avgLineHeight;

/** Text line width variance divided by line height mean */
   double _normLineWidthVar;

/** Text line left indent variance divided by left indent mean */
   double _normLineLeftIndentVar;

/** Text line right indent variance divided by right indent mean */
   double _normLineRightIndentVar;

};

#endif




