#ifndef PAIRSHAPE_H_HM_2005
#define PAIRSHAPE_H_HM_2005

#include <fstream>

#include "../PageLayoutLib/objectInfoClass.h"
#include "math.h"
#include <string>

#include "../Utility_hm/TMatrix.h"


/** The pairshape_vecotr data structure describes the shape information of 
 *   a quadrilateral (4 sided polygon) formed by two textlines.
 *   Five parameters are sufficiant to define a unique quadrilateral.
 *   Two options : 5 lengths or 3 lengths + 2 angles.
 *   The former is chosen as the latter brings normalization problem when 
 *   computing the similarity between two vectors, which is formulated as the
 *   Euclidean distance between two vectors.
 */

struct pairshape_vector
{
   double Len1;   // length of one textline
   double Len2;   // length of the other textline
   double Len3;   // length of the line which connect the midpoints of the two textlines
   double alpha;  // Length Between left endpoint of textline1 and right endpoint of textline2;
   double belta;  // Length Between right endpoint of textline1 and left endpoint of textline2;

   int		type1;			// type of the first object: textline, tableline, etc.
   int		type2;
   int		FontHeight1;	// font size of the first textline
   int		FontHeight2;	// font size of the second textline

   int		doc_ID; // reserved

   /** 
    * write_stream outputs instance to file
    * \param f output file stream object
    **/
   void write_stream( ofstream &f )
   {
      f << "---------pairshape-vector---------" << endl;

      f << Len1 << endl;   // length of one textline
      f << Len2 << endl;   // length of the other textline
      f << Len3 << endl;   // length of the line which connect the midpoints of the two textlines
      f << alpha << endl;  // Length Between left endpoint of textline1 and right endpoint of textline2;
      f << belta << endl;  // Length Between right endpoint of textline1 and left endpoint of textline2;

      f << type1 << endl;			// type of the first object: textline, tableline, etc.
      f << type2 << endl;
      f << FontHeight1 << endl;	// font size of the first textline
      f << FontHeight2 << endl;	// font size of the second textline

      f << doc_ID << endl; // reserved
	
   }

   /** 
    * read_stream loads instance from file
    * \param f input file stream object
    **/
   void read_stream( ifstream &f )
   {
      char	buf[100];
      f >> buf;

      f >> Len1 ;   // length of one textline
      f >> Len2 ;   // length of the other textline
      f >> Len3 ;   // length of the line which connect the midpoints of the two textlines
      f >> alpha ;  // Length Between left endpoint of textline1 and right endpoint of textline2;
      f >> belta ;  // Length Between right endpoint of textline1 and left endpoint of textline2;

      f >> type1 ;			// type of the first object: textline, tableline, etc.
      f >> type2 ;
      f >> FontHeight1 ;	// font size of the first textline
      f >> FontHeight2 ;	// font size of the second textline

      f >> doc_ID ;      // reserved

   }


   /** 
    * Constructor
    **/
   pairshape_vector()
   { 
      Len1 = Len2 = Len3 = alpha = belta = 0;

      FontHeight1 = 0;
      FontHeight2 = 0;

      type1 = type2 = -1;
      doc_ID = -1;
   }
	
   /** 
    * Constructor
    **/
   pairshape_vector(double l1, double l2, double l3, double a, double b)
   {
      Len1 = l1; Len2 = l2; Len3 = l3; alpha = a; belta = b;
   }
	
   public:
   void Construction(objectInfo &obj, pointLocation & ref1, pointLocation & ref2);
	
   pairshape_vector Mirror();

   double Euc_dist(pairshape_vector &other);

   double Hur_dist(pairshape_vector &other);

   void Normalization();

   pairshape_vector Reverse();

   double match(pairshape_vector &a,bool upside = false);

   void Construction(struct objectInfo& O1, struct objectInfo& O2);	

   pairshape_vector operator = (pairshape_vector a);

   bool operator == (pairshape_vector a);

   double norm();

   double normsquare();

   pairshape_vector operator - (pairshape_vector a);

   pairshape_vector operator + (pairshape_vector a);

   pairshape_vector operator * (double a);

   pairshape_vector operator / (double a);

   pairshape_vector dotproduct(pairshape_vector a);

   pairshape_vector dotsqrt();
	
   pairshape_vector dotdivide(pairshape_vector a);
	
   pairshape_vector objsqrt();

};




struct pairshape_vector_list
{
   struct pairshape_vector * m_pVectorList;
   int m_nVectors;

   /** 
    * Constructor
    **/
   pairshape_vector_list(){ m_pVectorList = NULL; m_nVectors = 0; }
   void Construction(struct objectInfo *obj, int object_count,int documentId, int style);

   /** 
    * Destructor
    **/
   ~pairshape_vector_list()
   {
      if(m_pVectorList)
         delete []m_pVectorList;
   }
   public:
   //void Construction2(struct objectInfo *obj, int object_count, int documentId);	
	
   void Append(pairshape_vector_list *addon);
   void Reset();

   /** 
    * read_stream loads instance from file
    * \param f input file stream object
    **/
   void read_stream( ifstream &f )
   {
      char buf[100];
      f >> buf;
      f >> buf; f >> m_nVectors;

      if( m_pVectorList ) delete []m_pVectorList;
      m_pVectorList	= new pairshape_vector[m_nVectors];
      for( int i=0; i<m_nVectors; i++ )
         m_pVectorList[i].read_stream( f )  ;
   };

   /** 
    * write_stream outputs instance to file
    * \param f output file stream object
    **/
   void write_stream( ofstream &f )
   {
      f << "***********pairshape-vector-list************" << endl;
      f << "nbVector_in_list: " << m_nVectors << endl;

      int i = 0;
      for(; i<m_nVectors; i++)
         m_pVectorList[i].write_stream(f);
		
   };
};



class pairshape_cluster
{
   public:
   pairshape_vector_list	m_MemberList;	
   int						m_nMember;
   pairshape_vector		m_centroid;
   pairshape_vector		m_sigma;

   double              m_weight;
   double              m_avgNum_perdoc;
   double              m_sigmaNum_perdoc;

   /** 
    * Constructor
    **/
   pairshape_cluster()
      {
         m_weight = 1;
      }

   void write_stream( ofstream &f );
   void read_stream( ifstream &f );
   void Reset();
   pairshape_cluster operator = (pairshape_cluster a);
};


#define KMEANS  0
#define DYN     1


#define		range_vector		0.2
#define		range_cluster		0.1
#define		range_weightcluster 0.1
#define		range_score			0.2

#define     this_iterations		2

/*
// for 5D txtline shape vector
#define		range_vector		0.05
#define		range_cluster		0.03
#define		range_weightcluster 0.03
#define		range_score			0.1

#define     this_iterations		2
*/

class pairshape_cluster_list
{
   public:
   pairshape_cluster * m_pClusterList;
   int m_nClusters;

   /** 
    * Constructor
    **/
   pairshape_cluster_list()
      {
         m_pClusterList = NULL; 
         m_nClusters = 0;
      }
   /** 
    * Desturctor
    **/
   ~pairshape_cluster_list()
      {
         if(m_pClusterList)
            delete []m_pClusterList;
      }

   /** 
    * release memory and reset member values
    **/
   void Reset()
      {
         int i;
         for(i=0;i<m_nClusters;i++)
            m_pClusterList[i].Reset();

         delete []m_pClusterList;

         m_pClusterList = NULL;
         m_nClusters = 0;
      }
   public:
   double  ProbScoreAPage(pairshape_vector_list * pTest, bool upside = false, bool mirror = false);
   double	m_nAvgVperDoc;
   double	m_noiseVRatio;	
   int		m_nTotalVectors;

   void   ClusterStat(int nDoc=1, bool flag = false);
   double Probability(pairshape_vector_list *pTest);
   void   CheckSTD(pairshape_vector &targ);
   double Probability(pairshape_cluster_list * pTest);
	
   void   dyn_Clustering(pairshape_cluster_list **pClusterList, int nList, bool stat_flag = true, bool flag = false);
   void   dyn_Clustering(pairshape_vector_list * pMultiDocLists, int nDoc=1, bool flag = false);
   double ScoreAPage(pairshape_cluster_list *pTest,int method,bool upside = true);
   double ScoreOnClusterList(pairshape_cluster_list *pList, int method = 1);
	
   void   WeightingAgainst(pairshape_cluster_list *pOther);
   void   SetDefaultSigma(pairshape_vector &sigma);

   //void   kmeans_Clustering(pairshape_vector_list * pDocLists, int nDoc, int nTargetClusters);
   void   SaveToDisk(string outfile, int style, bool saveweight = true);
   void   LoadFromDisk( string infile , int & style);
};



#ifndef INCREAMENTAL_AMOUNT
#define INCREAMENTAL_AMOUNT 12
#endif

struct b_Cluster
{
//	int				centroid[2]; // index of centroid object/pair, page_index/member_index
//	double			sum_dist;    //sum of distance from each object pair to centroid pair
   CTMatrix<int>	memberlist;
   int				nMember;	
   pairshape_vector centroid;

   int				nTotalMember; //for clustering on clusters

   /** 
    * Constructor
    **/
   b_Cluster()
   {
//		sum_dist = 0;
      nMember = 0;
      nTotalMember = 0;
//		centroid[0] = centroid[1] = -1;
   }

   /** 
    * Destructor
    **/
   ~b_Cluster()
   {
      memberlist.Destruction();
   }	
	
   /** 
    * Add new one/ones into the member list; at the same time, update center
    * \param page_index page index of the new member(s)
    * \param member_index member index of the new member
    * \param newone the new member(s)
    * \nNewmember the number of new members
    **/
   void Add(int page_index, int member_index, pairshape_vector &newone, int nNewMember=1)
   {
      if(nMember == memberlist.GetRows())
      {
         CTMatrix<int> tmp = memberlist;
         memberlist.Construction(nMember + INCREAMENTAL_AMOUNT,2);
         int i,j;
         for(i=0;i<nMember;i++)
            for(j=0;j<2;j++)
               memberlist[i][j] = tmp[i][j];
      }

      memberlist[nMember][0] = page_index;
      memberlist[nMember][1] = member_index;

      pairshape_vector tmp = centroid*(double)nTotalMember + newone*(double)nNewMember;
      centroid = tmp/(double)(nTotalMember+nNewMember);

//		ASSERT(centroid.norm()>0);

      nMember ++;
      nTotalMember += nNewMember;
   }
};

#endif
