#include "../Layout_hm/layout.h"
#include "TestingClass.h"
#include "utility.h"
#include "../Utility_hm/TArray.h"

/**
 * Constructor
 * \param sourcedataset source data path
 * \param outputfile output filename to store testing results
 * \param method the algorithm used for testing
 **/
TestingClass::TestingClass(string sourcedataset, string outputfile, int method)
{
   m_sDataPath		= sourcedataset;
   m_outputFname	= outputfile;
   m_method		= method;
   m_nTestDoc		= 0;
   m_sortflag		= true;
   m_maxNbTxline   = 128;	
   m_neighbor		= NEAREST_NEIGHBOR;

   m_wanted_pairshape_cluster  	= NULL;
   m_wanted_objcluster				= NULL;
   m_wanted_objPairClass			= NULL;

   m_psv_style		= -1;
}

/**
 * Destructor
 **/
TestingClass::~TestingClass()
{
   if(m_wanted_pairshape_cluster)
      delete m_wanted_pairshape_cluster;
   if(m_wanted_objcluster)
      delete m_wanted_objcluster;
   if(m_wanted_objPairClass)
      delete m_wanted_objPairClass;
}

/**
 * Reset release memory and reset all members to their default values
 **/
void TestingClass::Reset()
{
   m_sDataPath		= "";
   m_outputFname	= "";
   m_method		= PAIR_VECTOR_BOTH;
   m_nTestDoc		= 0;
   m_sortflag		= true;
   m_maxNbTxline   = 128;
   m_neighbor		= NEAREST_NEIGHBOR;

   m_FnameList.clear();
}

/**
 * LoadTrainingResults read training results from file
 * \param trainrstfile the training result file name
 **/
void TestingClass::LoadTrainingResults(string trainrstfile)
{
   //parse input file name
   char tmp = trainrstfile[trainrstfile.rfind('_')+1];
   int tmpmethod = atoi(&tmp);

   if(m_method >= 0 && tmpmethod != m_method)
   {
      printf("training method and testing method do not match!");
      return;
   }

   switch(m_method)
   {
      case PAIR_GENERIC :			
         m_wanted_objPairClass = new objectPairClass;
         m_wanted_objPairClass->LoadFromDisk(trainrstfile); 
         break;
      case PAIR_VECTOR_BOTH:
      case PAIR_VECTOR_WANTED:
         m_wanted_pairshape_cluster = new pairshape_cluster_list;
         int style;
         m_wanted_pairshape_cluster->LoadFromDisk(trainrstfile, style); 
         if( m_psv_style == -1) //undefined
            m_psv_style = style;
         else if(m_psv_style != style)
         {
            printf ("training pairshape_vector formation style does not match that of testing!");
            Reset();
            return;
         }				

         break;
      case SINGLE_VECTOR_BOTH:
      case SINGLE_VECTOR_WANTED: 
         m_wanted_objcluster = new Clayout;
         m_wanted_objcluster->LoadClustersFromDisk(trainrstfile); break;
		
      default:  printf("unknown method!"); break;		
   }

	
}

/**
 * Run exacutes the testing process,LoadTrainingResult() should be called before this
 **/
void TestingClass::Run()
{
   if(m_sDataPath.empty()) return;
	
   m_nTestDoc = LoadTiffImageSet( m_FnameList, m_sDataPath);

   if(m_nTestDoc == 0) 
   {
      printf("no testing images found!");
      return;
   }

   switch(m_method)
   {
      case PAIR_GENERIC :			RunPairGeneric(); break;
      case PAIR_VECTOR_BOTH:
      case PAIR_VECTOR_WANTED:	RunPairVector(); break;
      case SINGLE_VECTOR_BOTH:
      case SINGLE_VECTOR_WANTED:  RunObjVector(); break;
      default:  printf("unknown method!"); break;		
   }
	
   //output ranking results into files
   FILE * fp = fopen(m_outputFname.c_str(),"w");
   if(m_sortflag)
      SortScores(m_FnameList,m_scoreList, fp);
   else if(fp)
   {
      int i;
      for( i=0; i<m_nTestDoc; i++)
         fprintf(fp,"%s :   %.5f\n", m_FnameList[i],m_scoreList[i]);
   }
   fclose(fp);

}

/**
 * MyCmp compare two values
 **/
float * testscores;
int MyCmp(const void * arg1, const void * arg2)
{
   if(testscores[*(int *)arg1] > testscores[*(int *)arg2]) return -1;
   if(testscores[*(int *)arg1] < testscores[*(int *)arg2]) return 1;

   return 0;

}

/**
 * SortScores rank the documents and theirs scores in descending order
 * \param source the testing filename array
 * \param scorearray the array storing the score of each document
 * \param fp the output file pointer
 **/
void TestingClass::SortScores(cstring_array & source, CTArray<double> & scorearray, FILE *fp)
{
   if(scorearray.IsNull()) return;

   int size = source.size();
   int* index = new int[size];
	
   if(testscores) delete testscores;
   testscores = new float[size];

   int i;
   for( i=0;i<size;i++)
   {
      testscores[i] = scorearray[i];
      index[i] = i;
   }

   qsort(index, size, sizeof(int), MyCmp);

   if(fp)
      for(i=0;i<size; i++)
      {
         int ii = index[i];
         fprintf(fp,"%s :   %.8f\n", source[ii].c_str(),testscores[ii]);
      }

   delete testscores;
   testscores = NULL;

}

/**
 * RunPairGeneric exacutes the testing process using Arkin's turning functions for 
 * quadrilateral representation and distance computation
 * It sums the individual scores of polygons from testing document and normalize it by the total number of testing polygons. 
 * The score is computed based on neighbors in a predefined range around a testing polygon (cluster).
 * If using BEST_NEIGHBOR, the neighbor with the highest training weight is chosen and its weight used as the score;
 * NEAREST_NEIGHBOR uses the weight of the training neighbor that is most similar to the testing one
 * GAUSS_NEIGHBORHOOD uses the similarity weighted score of all neighbors as the testing score
 **/
void TestingClass::RunPairGeneric()
{
   if(m_wanted_objPairClass == NULL) return;

   m_scoreList.Construction(m_nTestDoc);
   int i;
   for( i=0;i<m_nTestDoc;i++)
   {
      objectPairClass * Test_pair = ExtractObjPair(m_FnameList[i], m_maxNbTxline);
      float tmpscore;
      m_wanted_objPairClass->scoreImage(Test_pair, &tmpscore, m_neighbor);
      m_scoreList[i] = (double) tmpscore;
   }

}

/**
 * RunPairVector computes the score of a testing document in a similar way as RunPairGeneric,
 * the only difference is due to the different representation and the distance computation.
 **/
void TestingClass::RunPairVector()
{	
   if(m_wanted_pairshape_cluster == NULL) return;

   m_scoreList.Construction(m_nTestDoc);
   int i;
   for( i=0; i<m_nTestDoc; i++)
   {
      pairshape_cluster_list *tmplist = ExtractObjPairVector(m_FnameList[i], m_psv_style,m_maxNbTxline, i);
      m_scoreList[i] = m_wanted_pairshape_cluster->ScoreAPage(tmplist, m_neighbor, true);
   }
}

/**
 * RunObjVector computes the score of a document as the product of inter-cluster probablity 
 * and intra-cluster probability. Inter-cluster probability captures the histogram distance(x-square)
 * distance,  whose bins hold the number of testing textlines that hit a specific training cluster; 
 * each bin being weighted by training cluster's weights.The probablity is exp(-dist*dist/sum_weights). 
 * Intra-cluster probability computes each bin's average hitting accuracy and then sums these accuracy weighted 
 * by the weights of corresponding training clusters.
 **/
void TestingClass::RunObjVector()
{
   if(m_wanted_objcluster == NULL) return;

   m_scoreList.Construction(m_nTestDoc);
   int i;
   for( i=0; i<m_nTestDoc; i++)
   {		
      Clayout *pLayout = new Clayout;
	
      //extract objects from each training document
      objectInfoClass *objects = new objectInfoClass(m_FnameList[i].c_str(),NULL,
                                                     m_maxNbTxline, NULL,0,1,0);
      pLayout->ImportObjects(objects->objectStructure(), objects->getObjectCount());
      delete objects;	

      m_scoreList[i] = BelongProb(m_wanted_objcluster->m_pClusterList,m_wanted_objcluster->m_nbClusters, pLayout);
      delete pLayout;
   }
}

