#include <string>
#include "TrainingClass.h"
#include "../Layout_hm/layout.h"
#include "../PageLayoutLib/objectInfoClass.h"
#include "../PageLayoutLib/objectPairClass.h"
#include "../Utility_hm/TMatrix.h"

#define CLUSTER_RANGE_OBJ	0.1

/**
 * Constructor
 * \param wantedfiles the path for storing positive/wanted training documents
 * \param unwantedfiles the path for storing negative/unwanted training documents
 * \param wantedoutput output file name for storing training results for wanted samples
 * \param unwantedoutput output file name for storing training results for unwanted samples
 * \param method the algorithm used for training
 **/
TrainingClass::TrainingClass(char * wantedfiles, char* unwantedfiles, char* wantedoutput, char * unwantedoutput, int method)
{
   m_sWanted_input_path		= wantedfiles ;
   m_sUnwanted_input_path		= unwantedfiles ;
   m_sWanted_output_fname		= wantedoutput ;
   m_sUnwanted_output_fname	= unwantedoutput ;

   m_method					= method;

   m_mergy_style				= DYN_APPROXIMATED;
	
   m_nWantedDoc				= 0;
   m_nUnwantedDoc				= 0;

   m_maxNbTxline				= 128;
   m_nWantedObjCluster			= 12;
   m_nUnwantedObjCluster		= 20;
   m_maxIter					= 5;

   m_psv_style					= PSV_N_2;
}

/**
 * Destructor
 **/
TrainingClass:: ~TrainingClass()
{
}

/**
 * Reset release memory and reset all members to their default values
 **/
void TrainingClass::Reset()
{
   m_sWanted_input_path.clear() ;
   m_sUnwanted_input_path.clear() ;
   m_sWanted_output_fname.clear() ;
   m_sUnwanted_output_fname.clear() ;

   m_method					= PAIR_GENERIC;
   m_mergy_style				= DYN_APPROXIMATED;
	
   m_nWantedDoc				= 0;
   m_nUnwantedDoc				= 0;

   m_maxNbTxline				= 128;
   m_nWantedObjCluster			= 20;
   m_nUnwantedObjCluster		= 20;
   m_maxIter					= 5;

   m_WantedFnameList.clear();
   m_UnwantedFnameList.clear();
}


/**
 * MergePairs
 * In training session, after each document's object pairs has been pruned, ie. 
 * each document's object pairs has been clustered for that document, 
 * call MergePairs() to further clustering on existing clusters accross all documents.
 * for method PAIR_GENERIC use only
 * \param dest points to the objectPairClass instance that holds the merge result
 * \param source points to the objectPairClass instances to be mergyed
 * \param nDoc number of documents involved
 **/
void TrainingClass::MergePairs(objectPairClass *dest, objectPairClass** source, int nDoc)
{
   dest->_numPairs = 0;
   int i, total_count;
   for ( i=0, total_count = 0;i<nDoc; i++) 
   {
      int nPairs = source[i]->_numPairsAfterPrune;
      total_count += nPairs;
   }

   dest->_objectPairs = (objectPairInfo *) calloc(total_count, sizeof(objectPairInfo));
//	ASSERT(dest->_objectPairs);

   dest->_numPairs = total_count;

   int index = 0;
   for(i = 0 ; i<nDoc;i++)
   {
      int j;
      int nPairs = source[i]->_numPairsAfterPrune;
      for ( j=0;j<nPairs;j++)
         dest->_objectPairs[index++] = source[i]->_objectPairs[j];
   }

   dest->_numDoc = nDoc;

   dest->prune2(false, false);

}

/**
 * Run exacutes training process
 **/
void TrainingClass::Run()
{
   if(m_sWanted_input_path.empty()) return;
	
	
   m_nWantedDoc = LoadTiffImageSet(m_WantedFnameList ,m_sWanted_input_path);

   if(m_nWantedDoc == 0) 
   {
      printf("no tiff images find in wanted file folder!");
      return;
   }

   m_nUnwantedDoc = ( m_method >=3 || m_sUnwanted_input_path.empty() ) ? 
      0 : 
      LoadTiffImageSet(m_UnwantedFnameList ,m_sUnwanted_input_path);
	
   if(m_method <3 && m_nUnwantedDoc == 0) 
   {
      printf("no unwanted file found!");
      return;
   }


   // append method info as the file name suffix
   char buf[100];
   if(!m_sWanted_output_fname.empty())
   {
      string tmp = m_sWanted_output_fname;	
      sprintf(buf, "%s_%d.txt", tmp.substr(0,tmp.rfind('.')).c_str(), m_method);
      m_sWanted_output_fname = string(buf);
   }
   if(!m_sUnwanted_output_fname.empty())
   {
      string tmp = m_sUnwanted_output_fname;	
      sprintf(buf,"%s_%d.txt",  tmp.substr(0,tmp.rfind('.')).c_str(), m_method);
      m_sUnwanted_output_fname = string(buf);
   }

   switch(m_method){
      case PAIR_GENERIC :			RunPairGeneric(); break;
      case PAIR_VECTOR_BOTH:
      case PAIR_VECTOR_WANTED:	RunPairVRoutines(); break;
      case SINGLE_VECTOR_BOTH:
      case SINGLE_VECTOR_WANTED:  RunObjVRoutines(); break;
      default:  printf("unknown method!"); break;		
   }	

}

/**
 * RunPairGeneric exacutes the training process using Arkin's turning functions for 
 * quadrilateral representation and distance computation
 * It represents textline pairs as 4-sided polygons, and Distance between two polygons 
 * is computed by using Arkin's distance metric:
 * i.e., integration of point difference between the turning functions which represent polygons
 * In this method, both wanted and unwanted training samples are used, though one can derive his
 * own method that does not use unwanted training samples, as the following two provided.
 **/
void TrainingClass::RunPairGeneric()
{
   //dealing with wanted dataset
   objectPairClass  ** pPairsInDoc = new objectPairClass*[m_nWantedDoc];
   int j;
   for(j=0; j<m_nWantedDoc; j++)
      pPairsInDoc[j] = ExtractObjPair(m_WantedFnameList[j]);

   objectPairClass * pWantedPairList = new objectPairClass;

   if(m_mergy_style == FIRST_CENTROID)
   {
      int k;
      for( k=0; k< m_nWantedDoc; k++)
         pWantedPairList->merge(pPairsInDoc[k],k);
   }
   else
      MergePairs(pWantedPairList, pPairsInDoc, m_nWantedDoc);

   for(j=0; j<m_nWantedDoc; j++) delete pPairsInDoc[j];
   delete []pPairsInDoc;

   //dealing with unwanted dataset
   pPairsInDoc = new objectPairClass*[m_nUnwantedDoc];
   for(j=0; j<m_nUnwantedDoc; j++)
      pPairsInDoc[j] = ExtractObjPair(m_UnwantedFnameList[j]);

   objectPairClass * pUnWantedPairList = new objectPairClass;

   if(m_mergy_style == FIRST_CENTROID)
   {
      int k;
      for( k=0; k< m_nUnwantedDoc; k++)
         pUnWantedPairList->merge(pPairsInDoc[k],k);
   }
   else
      MergePairs(pUnWantedPairList, pPairsInDoc, m_nUnwantedDoc);

   for(j=0; j<m_nUnwantedDoc; j++) delete pPairsInDoc[j];
   delete []pPairsInDoc;

   // weighting wanted Pair clusters
   pWantedPairList->assignScore(pUnWantedPairList); // m_nWantedDoc, m_nUnwantedDoc);

   // save training results to disk
   pWantedPairList->SaveToDisk(m_sWanted_output_fname);
   pUnWantedPairList->SaveToDisk(m_sUnwanted_output_fname);

   delete pWantedPairList;
   delete pUnWantedPairList;
}

/**
 * RunPairVRoutines represents textline pairs as a 5D vector which capture the shape information
 * of the 4sided polygon. Distance between two polygons is the simple Euclidean distance.
 * Specifically, the 5D are: lengths of the two textlines, length between midpoints of the two textlines, 
 * and the two diagonal lengths.  All are normalized to [0,1].
 * PAIR_VECTOR_BOTH use both wanted and unwanted training samples, while 
 * PAIR_VECTOR_WANTED only use wanted training samples.
 **/
void TrainingClass::RunPairVRoutines()
{
   pairshape_cluster_list ** tmplists = new pairshape_cluster_list*[m_nWantedDoc];
   int i;
   for( i=0; i<m_nWantedDoc; i++)
      tmplists[i] = ExtractObjPairVector(m_WantedFnameList[i], m_psv_style, m_maxNbTxline, i);
	
	
   pairshape_cluster_list * pWantedPairList = new pairshape_cluster_list;
   pWantedPairList->dyn_Clustering(tmplists, m_nWantedDoc, false, false);	
			
   for(i=0; i<m_nWantedDoc; i++) delete tmplists[i];
   delete []tmplists;
   tmplists = NULL;
		
		
   if(m_method == PAIR_VECTOR_BOTH)
   {
      tmplists = new pairshape_cluster_list*[m_nUnwantedDoc];	
      int j;
      for( j=0; j<m_nUnwantedDoc; j++)
         tmplists[j] = ExtractObjPairVector(m_UnwantedFnameList[j], m_psv_style, m_maxNbTxline, j);
		
      pairshape_cluster_list * pUnWantedPairList = new pairshape_cluster_list;		
      pUnWantedPairList->dyn_Clustering(tmplists, m_nUnwantedDoc, false, false);	
		
      for(i=0; i<m_nUnwantedDoc; i++) delete tmplists[i];
      delete []tmplists;

      pWantedPairList->WeightingAgainst(pUnWantedPairList);

      pWantedPairList->SaveToDisk(m_sWanted_output_fname, m_psv_style);
      pUnWantedPairList->SaveToDisk(m_sUnwanted_output_fname, m_psv_style, false);

      delete pUnWantedPairList;
   }

   else pWantedPairList->SaveToDisk(m_sWanted_output_fname, m_psv_style, false);
		
   delete pWantedPairList;	
	
}

/**
 * RunObjVRoutines() represents a document layout as a set of textline clusters. I.e, it uses 
 * single textline as its processing element, not textline pairs.
 * Each textline is represented as a 5D vector(length, position, font-height, angle from the horizontal line)
 * Distance is the Euclidean distance between two vectors.
 * SINGLE_VECTOR_BOTH use both wanted and unwanted training samples, while 
 * SINGLE_VECTOR_WANTED only use wanted training samples.
 **/
void TrainingClass:: RunObjVRoutines()
{
   Clayout *pLayout = new Clayout[m_nWantedDoc];
   int i;
   int sum_obj = 0;
   for( i=0; i<m_nWantedDoc; i++)
   {
      //extract objects from each training document
      objectInfoClass objects (m_WantedFnameList[i].c_str(),NULL,m_maxNbTxline, NULL,0,1,0);
      sum_obj += objects.getObjectCount();

      pLayout[i].ImportObjects(objects.objectStructure(), objects.getObjectCount());
//		delete objects;	
   }

   Clayout pLayout_wanted;
   objvector avg,sigma;
   pLayout_wanted.NormalizeObjects(pLayout, m_nWantedDoc, avg, sigma);			
   pLayout_wanted.KmeansObjClustering(pLayout,sum_obj,m_nWantedDoc, 
                                      m_nWantedObjCluster, 5, sigma, false);

   if(m_method == SINGLE_VECTOR_BOTH)
   {
      Clayout *pLayout2 = new Clayout[m_nUnwantedDoc];
      int j;
      int sum_obj2 = 0;
      for( j=0; j<m_nUnwantedDoc; j++)
      {
         //extract objects from each training document
         objectInfoClass objects(m_UnwantedFnameList[j].c_str(),NULL,m_maxNbTxline, 0,0,1,0);
         sum_obj2 += objects.getObjectCount();

         pLayout2[j].ImportObjects(objects.objectStructure(), objects.getObjectCount());
         //delete objects;	
      }

      Clayout pLayout_unwanted;
      objvector avg2,sigma2;
      pLayout_unwanted.NormalizeObjects(pLayout2, m_nUnwantedDoc, avg2, sigma2);			
      pLayout_unwanted.KmeansObjClustering(pLayout2,sum_obj2,m_nUnwantedDoc, 
                                           m_nUnwantedObjCluster, 5, sigma2, false);

      pLayout_wanted.Weight_Clusters_against( &pLayout_unwanted );

      pLayout_unwanted.SaveClustersToDisk(m_sUnwanted_output_fname, false);
      pLayout_wanted.SaveClustersToDisk(m_sWanted_output_fname);

   }
   else
      pLayout_wanted.SaveClustersToDisk(m_sWanted_output_fname, false);	

}


