#include "utility.h"
#include <string>
#include <stdio.h>

/**
 * FindFile finds all files with suffix specified by the input parameter
 * \param path the director to be searched
 * \param file_attr the filename suffix 
 * \filearray holds the filenames to be returned
 **/

#ifdef WIN32
#include <io.h>

void FindFile(string path, char * file_attr, cstring_array * filearray)
{
   struct _finddata_t tempFind; 
   string tempFileFind = path+ "\\*.*";

   long handle  = _findfirst(tempFileFind.c_str(), &tempFind);
   bool IsFinded =  handle != -1;
   while(IsFinded)
   {
      IsFinded = _findnext(handle, &tempFind) !=-1;  //int --> bool?
      if( strcmp(tempFind.name, ".") != 0
          && strcmp(tempFind.name, "..") != 0)

      {
         string foundFileName = tempFind.name;
		
         if(tempFind.attrib & _A_SUBDIR)
         {
            string  subPath = path + "\\" + foundFileName;
            FindFile(subPath,file_attr, filearray);				
         }
         else
         {
            string tmp_attr(file_attr);
            transform (tmp_attr.begin(),tmp_attr.end(), tmp_attr.begin(), toupper);

            //tmp_attr.MakeUpper();
            if(foundFileName.find(file_attr)==-1 &&
               foundFileName.find(tmp_attr) == -1 ) continue;  //not tiff file
				
            filearray->push_back(path + "\\"+ foundFileName);
         }
      }
   }	
   _findclose(handle);
}

#else
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

string toUpper( string str ){
   string s;
   const char * pch = str.c_str();
   while( *pch )
      s += toupper( *pch  );

   return s;
}

void FindFile(string path, char * file_attr, cstring_array * filearray)
{
   DIR             *dip;
   struct dirent   *dit;

   int             i = 0;

   if ((dip = opendir(path.c_str())) == NULL)
      perror("opendir");

   /*  struct dirent *readdir(DIR *dir);
    * Read in the files from given path and print to filearray */
   while ((dit = readdir(dip)) != NULL)
   {
      struct stat this_stat;
      stat( dit->d_name, & this_stat );

      string current_fname = dit->d_name;
      if( strcmp( dit->d_name, "." ) != 0 && strcmp( dit->d_name,".." )!= 0 )
      {
         if( S_ISREG( this_stat.st_mode )  ){
            string tmp_attr(file_attr);
            tmp_attr = toUpper( file_attr  );
		
            //tmp_attr.MakeUpper();
            if( current_fname.find(file_attr)==-1 &&
                current_fname.find(tmp_attr) == -1 ) continue;  //not tiff file

            filearray->push_back(path + "\\"+ current_fname );
         }
         else if( S_ISDIR( this_stat.st_mode ) )
         {
            string  subPath = path + "\\" + current_fname;
            FindFile(subPath,file_attr, filearray);
         }
      }	
   }
        
   closedir(dip);

}

#endif


/**
 * LoadTiffImageSet load all tiff image file names
 * \param retSet the returning array holding all the tiff filenames
 * \param filefolder the source file folder to be searched
 **/
int LoadTiffImageSet(cstring_array &retSet, string filefolder)
{
   FindFile(filefolder, ".tif", &retSet);
   return retSet.size();
}

/**
 * ExtractObjPair extract texline pairs from a tiff file, for method PAIR_GENERIC
 * \param fname the tiff file name
 * \param maxNbTxline the maximum number of text lines to be extracted
 **/
objectPairClass* ExtractObjPair(string fname, int maxNbTxline)
{
   objectInfoClass *objectInfo = new objectInfoClass( fname.c_str(),NULL,  //without using Jscript
                                                      maxNbTxline, NULL,0,1,0);
		
   objectPairClass *objectPairs = new objectPairClass( objectInfo->objectStructure(),
                                                       objectInfo->getObjectCount(),1,NULL); 
			
   delete objectInfo;

   return objectPairs; 
}


/**
 * ExtractObjPairVector extract texline objects and form pairvectors from a tiff file, for method PAIR_VECTOR_xx
 * \param fname the tiff file name
 * \param style specifies the object pair vector formation style
 * \param maxNbTxline the maximum number of text lines to be extracted
 * \param docID the index of the tiff file
 **/
pairshape_cluster_list* ExtractObjPairVector(string fname, int style, int maxNbTxline, int docID)
{
   objectInfoClass objectInfo (fname,NULL,   //Jscript surpressed
                               maxNbTxline, 0,0,1,0);	
 
   pairshape_vector_list *tmp = new pairshape_vector_list;

   tmp->Construction(objectInfo.objectStructure(),	
                     objectInfo.getObjectCount(),docID, style);
	
//	delete objectInfo;

   pairshape_cluster_list * ret = new pairshape_cluster_list;
   ret->dyn_Clustering(tmp,1);
   delete tmp;

   return ret;
}

/** 
 * BelongProb compute the probability of a testing layout belonging to a given layout class
 * \param pClusters points to a reference layout class ( a cluster set)
 * \param nClusters the number of clusters in the reference layout
 * \param pTest points to the testing layout
 **/
double BelongProb(objcluster *pClusters, int nCluster, Clayout *pTest)
{
   int nTestObj = pTest->m_nbObjects;

   CTArray<int>	 hit_histo(nCluster);
   CTArray<double>  hit_score(nCluster);

   int unhitted_obj=0;
   int i;
   for(i=0;i<nTestObj;i++)
   {
      objvector &obj = pTest->m_pObjectList[i];

      double maxprob = 0; 
      int maxindex = -1;
      int j;
      for(j=0;j<nCluster;j++)
      {
         // look for the cluster that current object hit
         double prob = pClusters[j].Probability(obj);
         if(prob > 0.1 && prob> maxprob)
         {
            maxprob = prob;
            maxindex = j;
         }
      }

      if(maxindex >=0) // some cluster is hit
      {
         hit_histo[maxindex] ++;			 // record how many objects hit the same cluster
         hit_score[maxindex] += maxprob;  // record the average hitting score
      }
      else
         unhitted_obj ++;
   }

   //use x^2 statistics to measure histogram difference
   double x_square = 0;
   double totalWeights = 0;
   for(i=0;i<nCluster;i++)
   {
      x_square +=   (hit_histo[i] - pClusters[i].m_avgnbObjperdoc) 
         * (hit_histo[i] - pClusters[i].m_avgnbObjperdoc)
         / (2* pClusters[i].m_sigmaObjperdoc)
         * pClusters[i].m_weight;

      totalWeights += pClusters[i].m_weight;
   }

   x_square +=  (unhitted_obj - nTestObj*0.1)
      *(unhitted_obj- nTestObj*0.1)
      /(2.0 * 5) ; 
	
//	int nb_unhit_bin = unhitted_obj>0? 1:0 ;
   x_square /= totalWeights;

   double intra_cluster_diff = 0; // the sum of the score got for each individual cluster
   int n_hitCluster=0;
   double SumWeights = 0;
   for(i=0;i<nCluster;i++)
   {
      if(hit_histo[i] <1 ) continue;
      intra_cluster_diff += pClusters[i].m_weight * hit_score[i] / (double)hit_histo[i];
      n_hitCluster++;
      SumWeights += pClusters[i].m_weight;
   }

//	double final_prob = intra_cluster_diff/ (1 + x_square);  			
//	double final_prob = intra_cluster_diff/n_hitCluster * exp(-x_square);

   double final_prob = intra_cluster_diff/SumWeights 
      * exp(-x_square) // - fabs(m_nAvgVperDoc - nTestObj)/(m_nAvgVperDoc+nTestObj)); 
      * ((nTestObj - unhitted_obj)/(double)nTestObj) ;

   return final_prob;

}
