/**
 * Kmeans.cpp: 
 * implementation of the CKmeans class
 * also includes implentation of struct aCluster
 **/

#include "Kmeans.h"
#include <string>
#include <math.h>
#include <stdlib.h>
#include <time.h>


#define MAXPATTERN 1000
#define MAX_ITERATION 10
#define MIN_DIFF_RATIO 0.0005


/**
 * Constructor
 * \param nPat total number of patterns need to be clustered
 * \param patDimension dimension of each source pattern
 * \param size_vector vector dimension (really used) of each source pattern
 * \param dyn boolean variable indicating whether dynamically determine the number of cluster or not
 * \param numCluster number of targetted clusters
 **/
CKmeans::CKmeans(int nPat,int patDimension,int size_vector,bool dyn,int numCluster)
{
   NumPatterns   = nPat;
   PatVectorDim  = patDimension;
   SizeVector    = size_vector;
   DynNumCluster = dyn;
   NumClusters   = numCluster;

   if(NumPatterns >0)
      Pattern.Construction(NumPatterns,PatVectorDim);

   // set default domain for each dimension
   maxRange.Construction(SizeVector);
   int i;
   for( i=0;i<size_vector; maxRange[i++] = 1);

   constraint.Construction(SizeVector);
   for(i=0;i<SizeVector;constraint[i++] = 1e10);
}


/**
 * Destructor
 **/
CKmeans::~CKmeans()
{
   Pattern.Destruction();
}

//DEL int CKmeans::LoadPatterns(char *fname)
//DEL {
//DEL    FILE *InFilePtr;
//DEL  
//DEL    double x;
//DEL 	if((InFilePtr = fopen(fname, "r")) == NULL)
//DEL 		return FAILURE;
//DEL 	fscanf(InFilePtr, "%d", &NumPatterns);  // Read # of patterns
//DEL 	fscanf(InFilePtr, "%d", &SizeVector);   // Read dimension of vector
//DEL 	fscanf(InFilePtr, "%d", &NumClusters);  // Read # of clusters for K-Means
//DEL 	for (int i=0; i<NumPatterns; i++)  // For each vector
//DEL 	{        
//DEL 	   for (int j=0; j<SizeVector; j++) // create a pattern
//DEL 	   {       
//DEL 		  fscanf(InFilePtr,"%lg",&x);       // consisting of all elements
//DEL 		  Pattern[i][j]=x;
//DEL 	   } 
//DEL 	} 
//DEL 	printf("Input patterns:\n");
//DEL 	for (i=0; i<NumPatterns; i++) 
//DEL 	   printf("Pattern[%d]=(%2.3f,%2.3f)\n",i,Pattern[i][0],Pattern[i][1]);
//DEL 	  
//DEL 	printf("\n--------------------\n");
//DEL 	return SUCCESS;
//DEL }

/**
 * InitClusterCenter Initialize clusters
 * \param flag boolean variable indicating whether randomly generating initial cluster centers;
 * If false, random generation; if true, using the first k members as initial centers. 
 **/
void CKmeans::InitClusterCenter(bool flag)  
{
   int i;
   if(!DynNumCluster) 
   {
      for( i=0;i<NumClusters;i++)
         Cluster[i].Initiate(SizeVector);	

      if(!flag)
      {
         srand( (unsigned)time( NULL ) );  //random seed based on system time

         for (int i=0; i<NumClusters; i++)
         {
            int index = ( ((double)rand())/RAND_MAX ) * NumPatterns;
            Cluster[i].Member[0]	=  index;
            for (int j=0; j<SizeVector; j++) 
               Cluster[i].Center[j]=Pattern[index][j];      
         }
      }
      else
      {
         int i;
         int j;
         for ( i=0; i<NumClusters; i++)
            for( j=0;j<SizeVector;j++)
               Cluster[i].Center[j] = InitialValue[i][j];
      }
   }

   else     // the mean of all data become one cluster centroid, and the most distant smaple
      // becomes the centroid of a second cluster
   {
      int i;
      int j;
      for( i=0;i<MAXCLUSTER;i++)	Cluster[i].Initiate(SizeVector);	

      CTArray<double> sum(SizeVector);
      for( j=0;j<SizeVector;j++)
      {
         for( i = 0;i<NumPatterns;i++)
            sum[j] += Pattern[i][j];

         Cluster[0].Center[j] = sum[j]/NumPatterns;
      }

      double maxdist = 0;
      int index = -1;
      for(i=0;i<NumPatterns;i++) // get the farest sample
      {
         double dist = EucNorm(i,0);
         if(dist>maxdist)
         { 
            maxdist = dist;
            index = i;
         }
      }
      for(j = 0;j<SizeVector;j++)
         Cluster[1].Center[j] = Pattern[index][j];

   }// end else

}

/**
 * RunKMeans exacute the clustering process
 * \param maxiteration maximum number of iterations to be exacuted
 * \param flag a boolean variable to be passed to InitClusterCenter()
 **/
void CKmeans::RunKMeans(int maxiteration,bool flag)
{
   InitClusterCenter(flag);
   DistributeSamples(); 
   CalcNewClustCenters();
   CalcJe();

   bConverged = FALSE;
   runtime = 1;

   int maxtimes = maxiteration ==0 ? MAX_ITERATION : maxiteration;
   while(runtime < maxiteration)
   {
      DistributeSamples();
      bConverged = CalcNewClustCenters();
      double oldJe = Je;
      CalcJe();               
      if(Je < oldJe*0.99 && bConverged ) break;  
      runtime++;
   } 

   int i;
   for( i=0;i<NumClusters;i++) // just for some statistical info
      CalcSampleDev(i);
}


/**
 * EncNorm calculates Euclidean norm of vector difference 
 * \param p index of one pattern
 * \param c index of an existing cluster
 **/
inline double CKmeans::EucNorm(int p, int c)   
{
   double dist = 0;                          
											  
   for (int  i=0; i<SizeVector ;i++)
   {
      if(ConditionSatisfied(Cluster[c].Center[i]-Pattern[p][i],i))
         dist += (Cluster[c].Center[i]-Pattern[p][i])*(Cluster[c].Center[i]-Pattern[p][i]);
      else
         dist += 1e10;  // set INF distance, if spectial condition is violated
   }
   return dist;
}

/**
 * FindClosestCluster returns indx of cluster closest to the pattern whose index is pat
 * \param pat index of a pattern
 * \param dist distance between pattern pat and its closet cluster
 **/
int CKmeans::FindClosestCluster(int pat,double &dist)
{
   int ClustID=-1;
   double MinDist=9.9e+99, d;

   for (int i=0; i<NumClusters; i++) 
   {
      d=EucNorm(pat,i);
      if (d<MinDist) 
      {
         MinDist=d;
         ClustID=i;
      }
   } 

   dist = d;
   return ClustID;
}

/**
 * DistributeSamples distribute every pattern to its neareat cluster
 **/
void CKmeans::DistributeSamples()    
{									 
   int Clustid,MemberIndex;

   //Clear membership list for all current clusters
   for (int i=0; i<NumClusters;i++)
   {
      Cluster[i].NumMembers = 0;
//	   Cluster[i].Sum_dist = 0;
   }

   for (int pat=0; pat<NumPatterns; pat++) 
   {
      //Find cluster center to which the pattern is closest
      double dist;
      Clustid = FindClosestCluster(pat,dist);
      if(Clustid == -1) continue;
//	   {
//		   char buf[6];
//		   int result = ::AfxMessageBox("Wrong Cluster found for pattern:" + CString(itoa(pat,buf,10)) + ", Ignore?",MB_OKCANCEL);
//		   if(result == IDOK) continue;
//		   else return;
//	   }

      //post this pattern to the cluster
      MemberIndex=Cluster[Clustid].NumMembers;
      Cluster[Clustid].Add(pat);

      //compute Je
//	   Je += dist;	
   } 

//	Je /= NumPatterns;
}

/**
 * CalcNewClustCenters calculate new cluster center for each cluster 
 * after all patterns have been redistributed
 **/
BOOL  CKmeans::CalcNewClustCenters()
{
   int VectID;
   double tmp[MAXVECTDIM];

   BOOL	ConvFlag=TRUE;
   int i,j,k;							
   for ( i=0; i<NumClusters; i++)               //for each cluster
   { 
      for (j=0; j<SizeVector; tmp[j++]=0.0);   // clear workspace		  
		  
      for (j=0; j<Cluster[i].NumMembers; j++)      //traverse member vectors
      {
         VectID=Cluster[i].Member[j];
         for (k=0; k<SizeVector; k++)          //traverse elements of vector
            tmp[k] += Pattern[VectID][k]; 			// add (member) pattern elmnt into temp
      }  

      for ( k=0; k<SizeVector; k++)             //traverse elements of vector
      { 
         if(Cluster[i].NumMembers == 0)		//put the center outside the definition range
         {
            Cluster[i].Center[k] = maxRange[k]*100; 
            ConvFlag=FALSE;
            continue;
         }
			  
         tmp[k]=tmp[k]/(double)Cluster[i].NumMembers;
			 
         if(fabs(tmp[k]-Cluster[i].Center[k]) > MIN_DIFF_RATIO * maxRange[k])	     
            ConvFlag=FALSE;

         Cluster[i].Center[k]=tmp[k];
      } 
   } 

   return ConvFlag;
}

/**
 * Reset resets all working variables
 **/
void CKmeans::Reset() 
{
   Pattern.Destruction();
   maxRange.Destruction();
   constraint.Destruction();

   NumClusters  = 0;
   NumPatterns = 0;
   SizeVector  = 0;
   Je			= 0;
//	threshold_ratio  = 0.2;
   DynNumCluster = false;
   bUseInitialValue = false;

   int i;
   for( i=0;i<MAXCLUSTER;i++)
      Cluster[i].Reset();

}

/**
 * Constructor
 **/
CKmeans::CKmeans()
{
   NumClusters = NumPatterns = SizeVector  = 0;
   DynNumCluster = false; 
   bUseInitialValue = false;
//	threshold_ratio  = 0.2;
}

/**
 * SetDynamic set dynamic clustering flag
 * \param d boolean variable indicating whether dynamically determine number of cluster or not
 * \param nClusterNum sets the number of targeted clusters when d is set to false
 **/
void CKmeans::SetDynamic(BOOL d ,int nClusterNum)
{
   DynNumCluster = d;
   if(d==FALSE)
      NumClusters = (MAXCLUSTER<nClusterNum)?MAXCLUSTER:nClusterNum;
   else
      NumClusters = 2; 

}


/** 
 * CalcJe calculate the value of the judge function
 **/
inline void CKmeans::CalcJe()  // Je = Sum(inter-cluster distance)/Sum(intra-cluster distance), 
   // for simplicity, square distance is used.	
{	
   Je = 0;
   double total_inter_dist  = 0; 
   double total_intran_dist = 0; 
	
   int i;
   for( i=0;i<NumClusters;i++)
   {
      if(Cluster[i].NumMembers == 0 ) continue;

      int j;
      for(j=0;j<Cluster[i].NumMembers;j++)//compute square distance between each sample and its centroid
      {
         int VectID = Cluster[i].Member[j];
         double dist = 0;
         int k;
         for( k=0;k<SizeVector;k++)
            dist += (Pattern[VectID][k]-Cluster[i].Center[k]) *
               (Pattern[VectID][k]-Cluster[i].Center[k]);

         total_intran_dist += dist;
      }

      int p;
      for( p  = i+1 ;p<NumClusters;p++)//compute square distance between every two cluster centroid
      {
         if(Cluster[p].NumMembers == 0 ) continue;
         int k;
         for( k=0;k<SizeVector;k++)
            total_inter_dist += (Cluster[i].Center[k] - Cluster[p].Center[k])*(Cluster[i].Center[k] - Cluster[p].Center[k]);
      }
   }
	
   Je = total_inter_dist / total_intran_dist;
}


/**
 * CalcSampleDev calculates the standard sample deviation for a cluster
 * \param ClusterID the index of the cluster 
 **/
void CKmeans::CalcSampleDev(int ClustID)
{
   if (ClustID < 0) return;
   Cluster[ClustID].ComputeSampleDev(Pattern, def_stdDev);
}

/**
 * Initiate allocates memories for source patterns
 **/
void CKmeans::Initiate()
{
   if(NumPatterns ==0) return;
   Pattern.Construction(NumPatterns,PatVectorDim);
//	threshold_ratio  = 0.2;

}

/**
 * InsideRange returns a boolean variable indicating whether a value is inside a given range
 * \param x the value to be tested
 * \param ref_min the lower side of the range
 * \param ref_max the upper side of the range
 **/
inline BOOL  InsideRange(double x,double ref_min,double ref_max)
{
   return x>= ref_min && x<= ref_max;
}

/**
 * SetDefaultStdDev set default standear deviation for clusters
 * \param def the array which holds the default values for each cluster
 **/
void CKmeans::SetDefaultStdDev(CTArray<double> &def)
{
//	assert(SizeVector == def.GetDimension());

   def_stdDev = def;

}

/**
 * SetInitialValue sets initial parameters (center and deviation) for each cluster
 * \param Initial_array the array holds the initial values 
 **/
void CKmeans::SetInitialValue(const CTMatrix<double> &Initial_array)
{
   if(Initial_array.IsNull()) return;
   InitialValue.Construction(Initial_array);

}

/**
 * SetBound4EachDim sets value range for each dimension of source data
 * \param boundarray the array holds the boundary values 
 **/
void CKmeans::SetBound4EachDim(double *boundarray)
{
   maxRange.Construction(SizeVector);
   int i;
   for( i=0;i<SizeVector;i++)
      maxRange[i] = boundarray[i];
}

/**
 * ConditionSatisfied tests whether the spetial constraint on a dimension is satisfied
 * \param diff the value to be tested
 * \param iDem the index of the dimension to be verified
 **/
bool CKmeans::ConditionSatisfied(double diff, int iDem)
{
   return fabs(diff) < constraint[iDem];
}

/**
 * SetConstraint sets constraint on a dimension
 * \param Dim the index of the dimension to be constrained
 * \param c the maximum value allowed on this dimension
 **/
void CKmeans::SetConstraint(int Dim, double c)
{
   if(constraint.IsNull())
   {
      constraint.Construction(SizeVector);
      int i;
      for( i=0;i<SizeVector;i++)
         constraint[i] = 1e10;
   }

   constraint[Dim] = c;
}



///////////////////////////////////////////////////////////////////
//aCluster implementations
////////////////////////////////////////////////////////////////

/**
 * ComputeSampleDev calclates sample deviation for a data set 
 * \param pattern the data set
 * \param def_std array that holds the default standard deviation when the computed value is zero
 **/
void aCluster::ComputeSampleDev(CTMatrix<double> &pattern, CTArray<double> &def_std)
{

//	SizeVector = pattern.Get_width(); 

//	for(int k = 0 ; k< SizeVector; SDev[k++] = 0);

//	ASSERT(dim<MAXVECTDIM);
   if(NumMembers < 1)		return;
   if(Member.IsNull())     return;

   int dim;
   int i;
   for( dim=0;dim<SizeVector;dim++)
   {
      SDev[dim] = 0;
      for( i = 0 ;i<NumMembers;i++)
      {
         int index = Member[i];
         SDev[dim] += (pattern[index][dim]-Center[dim])*(pattern[index][dim]-Center[dim]);
      }
			
      SDev[dim] /= (double)NumMembers ;  // devide by n or by n-1 ?
      SDev[dim] = sqrt(SDev[dim]);

      // if standard divation = 0, set it to a default value
      if(SDev[dim] == 0)
         SDev[dim] = def_std[dim];
   }

}


/**
 * Initiate allocates memory for a dataset
 * \param sizevector the number of dimensions of the source data
 **/
BOOL aCluster::Initiate(int sizevector)
{
   SizeVector = sizevector;

   Member.Construction(MAXPATTERN);
   return TRUE;

}

/**
 * Reset release memory 
 **/
void aCluster::Reset()
{
   Member.Destruction();
}

/**
 * operator = overides assign operator 
 **/
const aCluster & aCluster::operator = (aCluster b)
{
   Initiate(b.SizeVector);
   int i;
   for( i=0;i<SizeVector;i++)
   {
      Center[i] = b.Center[i];
      SDev[i] = b.SDev[i];
   }

   NumMembers  = b.NumMembers;
   Member.Destruction();
   if(NumMembers ==0) return *this;
   Member.Construction(b.Member);

   return *this;
}

/**
 * Add adds a pattern to the current cluster
 *\param pat the pattern to be appended
 **/
void aCluster::Add(int pat)
{
   if( NumMembers % MAXPATTERN ==0 && NumMembers>0 )
   {
      CTArray<int> tmp(NumMembers + MAXPATTERN);
      int i;
      for( i=0;i<NumMembers;tmp[i] = Member[i++]);
      Member = tmp;
   }

   Member[NumMembers++] = pat;

}
