/* This program is used to demonstrate the capabilities of the getLines  
   class.  After reading a deskewed, oriented, machine print text image,  
   this class can automatically determine the location of text and return  
   the bounding box coordinates and other related information.  Note that  
   the image must be deskewed and oriented in order to receive meaningful  
   results.  
  
   One can set GAUNTLET to one to loop over several calls to the getLines  
   class, setting different parameters each time.  By setting GAUNTLET to  
   zero, getLines is only called once.  
  
   By specifying an output file, the original image with the text lines  
   superimposed is output.  Each text line bounding box also shows the  
   central line of text, a single line used to represent the location of the  
   text line.  
  
   Returns 0 if successful, 1 otherwise. */  
  
// DECLARE HEADER FILES  
#include "getLines.h"  
#include "DLLineOrientationCalculation.h"
#include "DLBitsPerPixelConverter.h"
#include <DLLogger.h>  
#include "DLVersions.h"
  
// DECLARE MACROS  
#define GAUNTLET 0  
// 0: run once  
// 1: loop over hundred of calls to getLines  
#define MAXCHAR 500  
// the number of characters to allocate in the output error string  
  
// DECLARE SUBROUTINES  
char *getRootName(char *longFile, int num);  
void printStats(getLines *objects);  
  
// MEAT AND POTATOES  
int main (int argc, char **argv) {  
   DLConnectedComponents emuCon;  
   getLines *objects;  
   DLImage emu;  
   char *fileName = NULL, errChar[MAXCHAR];  
   float densityCCLess, maxToMedianRatio;  
   int highCCHeight, highCCWidth, minAvgHeight, lowCCHeight, lowCCWidth,  
      run = 0;  

   string version (argv[1]);

   if (version == "-v" || version == "--version") {
      std::cout << getLines::dlGetVersionInfo() << endl;
      return 0;
   }
  
   try {  
      // usage  
      if ((argc != 2) && (argc != 3)) {  
         // dlLogger.dlLog(DL_INFO, "Invalid argument count: %d\n", argc);  
         throw DLException(DL_UNKNOWN_FILE_FORMAT_EXCEPTION,  
                           "\nUsage: getLines <input image file> <optional output image file>\n"  
                           "\"getLines\" reads in a single image file and locates the text in "  
                           "a machine print deskewed, oriented image file with dark text on a "  
                           "light background.  Text line boundaries are returned.  If an "  
                           "output image file is provided, the bounding boxes of the text "  
                           "superimposed on the input file are provided as output.\n\n"
			   "-v OR --version provide version information");  
      }  
  
      // read input  
      ifstream inFile(argv[1]);  
      if (!inFile) {  
         // check to see that the file name provided by the argument exists  
         sprintf(errChar, "Input file %s does not exist.\n", argv[1]);  
         throw DLException(DL_FILE_NOT_FOUND_EXCEPTION, errChar);  
      }  

      // extract text lines using default values  
      emu.dlLoadImage(argv[1]);  


      /**  line orientation example **

      if (DLLineOrientationCalculation::dlCalculateLineOrientation(emu) == DL_VERTICAL)
      {
      dlLogger.dlLog(DL_CRITICAL,  
      "%s Line Orientation Vertical, Rotating 90 degrees.\n", argv[1]);  
      emu = emu.dlRotateImage(90);
      } 
	
      */

      objects = new getLines(emu);  
      if (objects->dlGetInfo() != NULL) {  
         run = (GAUNTLET) ? run++ : run;  
         // only increment if running GAUNTLET  
         if (argc == 3) {  
            if ((fileName = getRootName(argv[2], run)) == NULL)  
               // extract the root file name, insert run, and make sure image  
               // format suffix is replaced.  be sure to free when done.  
               // this is required since docLib requires the suffix name   
               // to determine the image format for the output file.  
            {  
               throw DLException(DL_NULL_POINTER_EXCEPTION,  
                                 "Problem with getRootName().\n");  
            }  
            dlLogger.dlLog(DL_CRITICAL,  
                           "Text line bounding box output file for %s (run %d): %s\n",  
                           argv[1], run, fileName);  
            objects->dlPrintLineInfo((string) fileName);  
            free(fileName);  
         } else {  
            dlLogger.dlLog(DL_CRITICAL,  
                           "Text line bounding box info for %s (run %d):\n", argv[1], run);  
            objects->dlPrintLineInfo();  
         }  
         printStats(objects);  
      } else {  
         sprintf(errChar, "(run %d): Unable to process file with "  
                 "objects->dlGetInfo().\n", run);  
         throw DLException(DL_NULL_POINTER_EXCEPTION, errChar);  
      }  
  
      // run extract text lines again but this time use different parameters  
      if (GAUNTLET) {  
         for (densityCCLess = 0.6; densityCCLess < 1.0; densityCCLess += 0.2) {  
            objects->dlSetFilterCCDensity(densityCCLess);  
            for (lowCCHeight = 0; lowCCHeight < 6; lowCCHeight += 2) {  
               for (highCCHeight = 300; highCCHeight < 600; highCCHeight += 200) {  
                  objects->dlSetFilterCCHeightBetween(lowCCHeight, highCCHeight);  
                  for (lowCCWidth = 0; lowCCWidth < 6; lowCCWidth += 3) {  
                     for (highCCWidth = 300; highCCWidth < 600; highCCWidth += 200)  
                     {  
                        objects->dlSetFilterCCWidthBetween(lowCCWidth, highCCWidth);  
                        for (maxToMedianRatio = 4.0; maxToMedianRatio < 8.0;  
                             maxToMedianRatio += 2.0)  
                        {	        
                           objects->dlSetMaxToMedianRatio(maxToMedianRatio);  
                           for (minAvgHeight = 5; minAvgHeight < 9; minAvgHeight += 2)  
                           {  
                              objects->dlSetMinAvgHeight(minAvgHeight);  
                              objects->dlRecalculate();  
                              // calling this method re-computes text line  
                              // information but with the new parameters invoked  
                              // called by the set methods  
                              if (objects->dlGetInfo() != NULL) {  
                                 run++;  
                                 dlLogger.dlLog(DL_CRITICAL,  
                                                "Run %d set method parameters: ", run);  
                                 dlLogger.dlLog(DL_CRITICAL,  
                                                "densityCCLess: %f lowCCHeight: %d ", densityCCLess,  
                                                lowCCHeight);  
                                 dlLogger.dlLog(DL_CRITICAL, "highCCHeight: ");  
                                 dlLogger.dlLog(DL_CRITICAL,  
                                                "%d lowCCWidth: %d highCCWidth: %d ", highCCHeight,  
                                                lowCCWidth, highCCWidth);  
                                 dlLogger.dlLog(DL_CRITICAL,  
                                                "maxToMedianRatio: %f minAvgHeight: %d\n",  
                                                maxToMedianRatio, minAvgHeight);  
                                 if (argc == 3) {  
                                    if ((fileName = getRootName(argv[2], run)) == NULL) {  
                                       throw DLException(DL_NULL_POINTER_EXCEPTION,  
                                                         "Problem with getRootName().\n");  
                                    }  
                                    dlLogger.dlLog(DL_CRITICAL,  
                                                   "Text line bounding box output file for ");  
                                    dlLogger.dlLog(DL_CRITICAL, "%s (run %d): %s\n",  
                                                   argv[1], run, fileName);  
                                    objects->dlPrintLineInfo((string) fileName);  
                                    free(fileName);  
                                 } else {  
                                    dlLogger.dlLog(DL_CRITICAL,  
                                                   "Text line bounding box info for %s (run %d):\n",  
                                                   argv[1], run);  
                                    objects->dlPrintLineInfo();  
                                 }  
                                 printStats(objects);  
                              } else {  
                                 sprintf(errChar, "(run %d): Unable to process "  
                                         "file with objects->dlGetInfo().\n", run);  
                                 throw DLException(DL_NULL_POINTER_EXCEPTION, errChar);  
                              }  
                           }  
                        }  
                     }  
                  }  
               }  
            }  
         }  
      }  
      delete objects;  
   } catch(DLException &Ex) {  
      dlLogger.dlLog("\nERROR in getLines main(): ", DL_CRITICAL);  
      dlLogger.dlLog(Ex.dlGetErrorMsg(), DL_CRITICAL);  
      return(1);  
   }  
   return(0);  
}  
  
/***************************************************************************  
 SUBROUTINES  
****************************************************************************/  
  
/***************************************************************************  
 *  Function:  getRootName  
 *  Description:  Given a filename with or without concatenated path,  
 *                creates the root filename by inserting the numeric  
 *                extension  
 *  Returns:  a filename if successful, NULL otherwise  
 ***************************************************************************/  
  
char *getRootName(char *longFile, int num) {  
   char *moa, *rootName, *suf;  
   int rootLen;  
   static char *outName = NULL;  
  
   // allocate memory for the root name and the suffix  
   rootLen = strlen(longFile);  
   try {  
      if ((rootName = (char *) calloc(rootLen + 1, sizeof(char))) == NULL) {  
         throw DLException(DL_MEMORY_EXCEPTION,  
                           "Unable to allocate memory for rootName.\n");  
      }  
      if ((suf = (char *) calloc(rootLen + 1, sizeof(char))) == NULL) {  
         throw DLException(DL_MEMORY_EXCEPTION,  
                           "Unable to allocate memory for suf.\n");  
      }  
  
      // get the file extension from the longFile  
      (void) strcpy(rootName, longFile);  
      moa = longFile + strlen(longFile) - 1;  
      while(moa > longFile) {  
         rootLen--;  
         if (*moa == '.') {  
            (void) strcpy(suf, moa + 1);  
            break;  
         }  
         moa--;  
      }  
      if (moa == longFile) {  
         suf[0] = '\0';  
      }  
  
      // this says where the root ends  
      rootName[rootLen] = '\0';  
  
      // create the output file name  
      if (num) {  
         if ((outName = (char *) calloc(strlen(rootName) + strlen(suf) + 7 +  
                                        num/10, sizeof(char))) == NULL)  
         {  
            throw DLException(DL_MEMORY_EXCEPTION,  
                              "Unable to allocate memory for outName.\n");  
         }  
         sprintf(outName, "%s.%04d.%s", rootName, num, suf);  
      } else {  
         if ((outName = (char *) calloc(strlen(rootName) + strlen(suf) + 2,  
                                        sizeof(char))) == NULL)  
         {  
            throw DLException(DL_MEMORY_EXCEPTION,  
                              "Unable to allocate memory for outName.\n");  
         }  
         sprintf(outName, "%s.%s", rootName, suf);  
      }  
   } catch(DLException &Ex) {  
      dlLogger.dlLog("\nERROR in getRootName() ", DL_CRITICAL);  
      dlLogger.dlLog(Ex.dlGetErrorMsg(), DL_CRITICAL);  
      return(NULL);  
   }  
  
   // free memory for temporary strings  
   free(rootName);  
   free(suf);  
  
   return(outName);  
}  
  
/***************************************************************************/  
  
void printStats(getLines *objects) {  
   int index, lineCount;  
   DLlineInfo *textLines;  
  
   lineCount = objects->dlGetLineCount();  
   dlLogger.dlLog(DL_CRITICAL, "STATS:\n");  
   if (lineCount) {  
      dlLogger.dlLog(DL_CRITICAL,  
                     "Text lines are bordered by upper left and lower right ");  
      dlLogger.dlLog(DL_CRITICAL,  
                     "coordinates (%d, %d) and (%d, %d), respectively.\n",  
                     objects->dlGetMinTextCol(), objects->dlGetMinTextRow(),  
                     objects->dlGetMaxTextCol(), objects->dlGetMaxTextRow());  
      dlLogger.dlLog(DL_CRITICAL, "Smallest text line height: %d\n",  
                     objects->dlGetMinTextHeight());  
      dlLogger.dlLog(DL_CRITICAL, "Smallest text line width: %d\n",  
                     objects->dlGetMinTextWidth());  
      dlLogger.dlLog(DL_CRITICAL, "Smallest text line y-intercept: %f\n",  
                     objects->dlGetMinIntercept());  
      dlLogger.dlLog(DL_CRITICAL, "Smallest text line slope: %f\n",  
                     objects->dlGetMinSlope());  
      dlLogger.dlLog(DL_CRITICAL, "Largest text line height: %d\n",  
                     objects->dlGetMaxTextHeight());  
      dlLogger.dlLog(DL_CRITICAL, "Largest text line width: %d\n",  
                     objects->dlGetMaxTextWidth());  
      dlLogger.dlLog(DL_CRITICAL, "Largest text line y-intercept: %f\n",  
                     objects->dlGetMaxIntercept());  
      dlLogger.dlLog(DL_CRITICAL, "Largest text line slope: %f\n",  
                     objects->dlGetMaxSlope());  
      dlLogger.dlLog(DL_CRITICAL, "Average text line height: %f\n",  
                     objects->dlGetAvgHeight());  
      dlLogger.dlLog(DL_CRITICAL, "Average text line width: %f\n",  
                     objects->dlGetAvgWidth());  
      dlLogger.dlLog(DL_CRITICAL, "Average text line slope: %f\n",  
                     objects->dlGetAvgSlope());  
      dlLogger.dlLog(DL_CRITICAL, "Text height population variance: %f\n",  
                     objects->dlGetTextVarianceHeight());  
      dlLogger.dlLog(DL_CRITICAL, "Text width population variance: %f\n",  
                     objects->dlGetTextVarianceWidth());  
   } else {  
      dlLogger.dlLog(DL_CRITICAL, "No text lines found.\n");  
   }  
  
   // access individual text line information  
   textLines = objects->dlGetInfo();  
   for (index = 0; index < lineCount; index++) {  
      dlLogger.dlLog(DL_CRITICAL, "Text line %d:\n", index + 1);  
      dlLogger.dlLog(DL_CRITICAL, "  Upper left coordinates: (%d, %d)\n",  
                     textLines[index].loc[0].x, textLines[index].loc[0].y);  
      dlLogger.dlLog(DL_CRITICAL, "  Lower right coordinates: (%d, %d)\n",  
                     textLines[index].loc[1].x, textLines[index].loc[1].y);  
      dlLogger.dlLog(DL_CRITICAL, "  Central left coordinates: (%d, %d)\n",  
                     textLines[index].loc[2].x, textLines[index].loc[2].y);  
      dlLogger.dlLog(DL_CRITICAL, "  Central right coordinates: (%d, %d)\n",  
                     textLines[index].loc[3].x, textLines[index].loc[3].y);  
      dlLogger.dlLog(DL_CRITICAL, "  Upper extent: %f\n",  
                     textLines[index].upperExt);  
      dlLogger.dlLog(DL_CRITICAL, "  Lower extent: %f\n",  
                     textLines[index].lowerExt);  
      dlLogger.dlLog(DL_CRITICAL, "  Pixel height: %d\n",  
                     textLines[index].height);  
      dlLogger.dlLog(DL_CRITICAL, "  Pixel width: %d\n",  
                     textLines[index].width);  
      dlLogger.dlLog(DL_CRITICAL, "  y-intercept: %f\n",  
                     textLines[index].yIntercept);  
      dlLogger.dlLog(DL_CRITICAL, "  Line slope: %f\n",  
                     textLines[index].slope);  
   }  
   dlLogger.dlLog(DL_CRITICAL, "\n\n");  
}  
  
/***************************************************************************/  
