/***************************************************************************/  
  
/*!\mainpage getLinesDOCLIB
  <br>Overview<br>
  <br>
  This class is used to identify and extract text lines from an image file.
  See
  <a href="./classgetLines.html#overview">getLines Class Reference: Detailed
  Description</a>
  for more information.<br>
  <br>
  <br>
  <br>Input Image Format<br>
  <br>
  The input image file should be a single image file of any format that the
  doclib library can handle.  Best results are achieved when the input image
  is a deskewed, oriented text image with dark machine print text on a light
  background.
*/

/**  
     \file getLines.h  
     \brief  
     getLines.h is the getLines header file.  It contains declarations of  
     the class, methods, variables, and structures required to perform text line  
     extraction.  It also contains most of the class documentation. */  
  
#ifndef __getLines_H  
#define __getLines_H  
  
#include <tiff.h>  
#include <DLConnectedComponents.h>  
  
using namespace std;  
  
const static  char DL_GETLINES_VERSION[] =  "getLinesDOCLIB Beta v1.0 - $Name:  $";


/***************************************************************************  
STRUCTURE DECLARATIONS  
***************************************************************************/  
  
/**   
      \struct compInfo  
      \brief A structure for storing connected component information. */  
typedef struct {  
   /** A number used to identify a connected component. */  
   int number;  
  
   /** The pixel height of a connected component. */  
   int height;  
} compInfo;  
  
/**   
      \struct lineList  
      \brief A structure for storing point information in a link list. */  
typedef struct {  
   /** The previous point. */  
   int prev;  
  
   /** The next point. */  
   int next;  
} lineList;  
  
/**   
      \struct pointLoc  
      \brief A structure to describe a point. */  
struct pointLoc {  
   /** The x-coordinate value of a point. */  
   int x;  
  
   /** The y-coordinate value of a point. */  
   int y;  
};  
  
/**   
      \struct DLlineInfo  
      \brief A structure that contains text line information. */  
typedef struct {  
   /** The number of connected components. */  
   int compCount;  
  
   /** The slope of the central text line. */  
   float slope;  
  
   /** The y-intercept of the central text line. */  
   float yIntercept;  
  
   /** The average (arithmetic mean) connected component pixel height. */  
   float avgHeight;  
  
   /** The average (arithmetic mean) connected component pixel width. */  
   float avgWidth;  
  
   /** The pixel height of the text bounding box. */  
   int height;  
  
   /** The pixel width of the text bounding box. */  
   int width;  
  
   /** The squared length of the central text line. */  
   float sqrLength;  
  
   /** The upper extent of the text line relative to line fit. */  
   float upperExt;  
  
   /** The lower extent of the text line relative to line fit. */  
   float lowerExt;  
  
   /** A list of the connected components. */  
   int *compList;  
  
   /** Flag to indicate if lines should be merged. */  
   int mergedFlag;  
  
   /** Text bounding box information:<br>  
       &nbsp;&nbsp;&nbsp;loc[0] are the upper left coordinates.<br>  
       &nbsp;&nbsp;&nbsp;loc[1] are the lower right coordinates.<br>  
       &nbsp;&nbsp;&nbsp;loc[2] are the central text line left coordinates.<br>  
       &nbsp;&nbsp;&nbsp;loc[3] are the central text line right coordinates.<br>  
   */  
   struct pointLoc loc[4];  
} DLlineInfo;  
  
/**   
      \struct DLnodes  
      \brief A structure for storing node information. */  
typedef struct {  
   /** Horizontal node weight. */  
   int hnode_wt;  
  
   /** Vertical node weight. */  
   int vnode_wt;  
  
   /** Horizontal "to" edge. */  
   int hto_edge;  
  
   /** Vertical "to" edge. */  
   int vto_edge;  
  
   /** Horizontal "from" edge. */  
   int hfrom_edge;  
  
   /** Vertical "from" edge. */  
   int vfrom_edge;  
} DLnodes;  
  
/**  
     \struct DLstats  
     \brief This structure is used by the calcHorizontalNeighbors method.  
     It stores line statistics. */  
typedef struct {  
   /** Horizontal mean. */  
   float hor_mean;  
  
   /** Vertical mean. */  
   float ver_mean;  
  
   /** [Sum(connected component maximum values -  
       connected component minimum values) + number of connected components]  
       / (number of connected components - 1). */  
   float asr;  
  
   /** Horizontal number.  Count of non-zero DLnodes hto_edge field  
       occurrences. */  
   float hor_num;  
  
   /** Vertical number.  Count of non-zero DLnodes vto_edge field  
       occurrences. */  
   float ver_num;  
  
   /** Average distances between column minimum and column maximum values when  
       the DLnodes hto_edge field occurrences are non-zero. */  
   float hor_dx;  
  
   /** Average distances between row minimum and row maximum values when the  
       DLnodes vto_edge field occurrences are non-zero. */  
   float ver_dy;  
} DLstats;  
  
/***************************************************************************  
CLASS DECLARATION  
***************************************************************************/  
  
/**  
     \class getLines  
     \brief This is the class for identifying and extracting text lines from an  
     image file.   
  
     <p><a name="overview"> </a>
     getLines.cpp contains the getLines class source code.  Given a  
     deskewed, oriented text image with dark machine print text on a light  
     background, this code will calculate text line information based on the  
     connected components.  The image must be deskewed and oriented in order to  
     receive meaningful results.  It was originally used in code called  
     "CollectFeatureData" which made use of subroutines from the TwoBit image  
     processing library.</p>  
  
     <p>The algorithm utilized by the getLines class has proven robust given  
     text images with low resolution, broken characters, connected scripts (e.g.  
     Arabic), multi-component character scripts (e.g. Chinese), and noise.</p>  
  
     <p>There are five basic steps to performing text line detection:<br>  
     &nbsp;&nbsp;&nbsp;Computing connected components.<br>  
     &nbsp;&nbsp;&nbsp;Filtering connected components based on size and  
     density.<br>  
     &nbsp;&nbsp;&nbsp;Computing line segment based on nearby neighbors.<br>  
     The centroids of the filtered connected components are used to form an  
     initial set of line segments.  This is done by connecting each component to  
     its nearest neighbor in a triangular region to the right while applying  
     distance constraints.<br>  
     &nbsp;&nbsp;&nbsp;Iteratively merging line segments to form new line  
     segments.  This is done based on the minimum distance between line  
     segments and horizontal and vertical line separation.  Merging continues  
     until no more merges occur.<br>  
     &nbsp;&nbsp;&nbsp;Final line segment filtering.</p>
  
     <p>In the below diagram, assume the skewed word "HELLO" has been correctly  
     identified as comprising a text line.  The central text line is defined by  
     endpoints {(x3, y3), (x4, y4)}.  The bounding box which encompasses "HELLO"  
     is defined by upper left coordinate (x1, y1) and lower right coordinate  
     (x2, y2).  The pixel height of the text line is y2-y1=9 and the pixel width  
     is x2-x3=38.  The density of the text line is 0.2135 which is 73 text pixels  
     divided by the bounding box area (9x38).  The slope of the central text line  
     is (y3-y4)/(x3-x4).  The y-intercept of the central text line is  
     y3-((y3-y4)/(x3-x4))*x3.  The upper extent is min(y3-y1, y4-y1) and the  
     lower extent is max(y3-y2, y4-y2).</p>  
     <DFN>  
     <br><br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(x1,&nbsp;y1)<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;v<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|--------------------------------------|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;####&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;#|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;####&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;#|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|#&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;##&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-------|<-(x4,&nbsp;y4)<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|#&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;#&nbsp;####&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;----------/&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;#|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|######&nbsp;&nbsp;##---------/&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;####&nbsp;|<br>  
     (x3,&nbsp;y3)->|---------/&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;&nbsp;##&nbsp;&nbsp;######&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|#&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;#&nbsp;####&nbsp;&nbsp;####&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|#&nbsp;&nbsp;&nbsp;&nbsp;#&nbsp;&nbsp;##&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|--------------------------------------|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;^<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
     &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(x2,&nbsp;y2)<br><br></DFN>  
  
     <p>Written by AMS on July 20, 2004 using code and documentation from LMG.</p>  
*/  
// The below diagram is the same as the above but not formatted for Doxygen.  
/*  
    (x1, y1)  
    |  
    v  
    |--------------------------------------|  
    |                                 #### |  
    |                        #       #    #|  
    |          ####  #       #       #    #|  
    |#    #  ##      #       #      -------|<-(x4, y4)  
    |#    #  # ####  #   ----------/ #    #|  
    |######  ##---------/    #        #### |  
    (x3, y3)->|---------/      #   ##  ######        |  
    |#    #  # ####  ####                  |  
    |#    #  ##                            |  
    |--------------------------------------|  
    ^  
    |  
    (x2, y2)  
*/  
class getLines {  
   public:  
   /***********************************************************************  
    VARIABLES  
   ***********************************************************************/  
  
   /** Text line information stored in an array of DLlineInfo structures.  
       Knowledge of _lineCount is required to access all terms. */  
   DLlineInfo *imageObj;  
  
   /***********************************************************************  
    CONSTRUCTOR AND DESTRUCTOR METHODS  
   ***********************************************************************/  
  
   /** 
    * Constructor that takes a DLImage as input.  Allocate memory and  
    * initialize variables.  
    * \param emu The DLImage of the image to process. 
    *        The getLines algorithm operates on binary image data.  
    *        If a color or grayscale image is passed in binarization 
    *        will be performed using the 
    *        DLBitsPerPixelConverter::dlDownscaleColor2Gray_global() and/or
    *        DLBitsPerPixelConverter::dlDownscaleGray2Binary_global() routines.
    *        If alternate binarization is desired it should be performed priof to invoking this class.
    *
    * 
    */  
   getLines(DLImage emu);  
  
   /** Destructor.  Clear allocated memory. */  
   ~getLines();  
  
   /***********************************************************************  
    REGULAR METHODS  
   ***********************************************************************/  
  
   /** Method to print text line information to standard output.  This  
       includes the upper right and lower left bounding box information. */  
   void dlPrintLineInfo();  
   void dlPrintLineInfo1(std::string fname, DLlineInfo *,int );  
  
   /** Method to print line information to standard output and create an  
       image called fname which contains the input image with text line  
       bounding boxes and central text lines superimposed.  
       \param fname Name of output image. */  
   void dlPrintLineInfo(std::string fname);  
  
   /** After setting new parameters with set methods, identify new text  
       lines using these parameters.  The only parameters that may be  
       adjusted are the following which are described in greater detail in  
       the Member Data Documentation section:<br>  
       &nbsp;&nbsp;&nbsp;_densityCCLess: Density upper threshold.<br>  
       &nbsp;&nbsp;&nbsp;_lowCCHeight: Lower pixel height threshold.<br>  
       &nbsp;&nbsp;&nbsp;_highCCHeight: Upper pixel height threshold.<br>  
       &nbsp;&nbsp;&nbsp;_lowCCWidth: Lower pixel width threshold.<br>  
       &nbsp;&nbsp;&nbsp;_highCCWidth: Upper pixel width threshold.<br>  
       &nbsp;&nbsp;&nbsp;_maxToMedianRatio: Upper height factor threshold.<br>  
       &nbsp;&nbsp;&nbsp;_minAvgHeight: Lower pixel height threshold.<br> */  
   void dlRecalculate();  
  
   /***********************************************************************  
    GET METHODS  
   ***********************************************************************/  
  
   /** Return the pixel height of the average (arithmetic mean) length line  
       of text.  
       \return float _avgHeightText */  
   float dlGetAvgHeight() const { return(_avgHeightText); }  
  
   /** Return the average (arithmetic mean) slope value for the central text  
       lines.  
       \return float _avgSlope*/  
   float dlGetAvgSlope() const { return(_avgSlope); }  
  
   /** Return the pixel width of the average (arithmetic mean) length line  
       of text.  
       \return float _avgWidthText*/  
   float dlGetAvgWidth() const { return(_avgWidthText); }  
  
   /** Return text line information in an array of DLlineInfo structures.  
       Knowledge of _lineCount is required to access all terms.  
       \return DLlineInfo imageObj*/  
   DLlineInfo *dlGetInfo() const { return(imageObj); }  
  
   /** Return the text line count.  
       \return int _lineCount*/  
   int dlGetLineCount() const { return(_lineCount); }  
  
   /** Return the largest y-ntercept value for the central text lines.  
       \return float _maxIntercept*/  
   float dlGetMaxIntercept() const { return(_maxIntercept); }  
  
   /** Return the largest slope value for the central text lines.  
       \return float _maxSlope*/  
   float dlGetMaxSlope() const { return(_maxSlope); }  
  
   /** Return the largest pixel column number which bounds text.  
       \return int _maxXText*/  
   int dlGetMaxTextCol() const { return(_maxXText); }  
  
   /** Return the pixel height of the tallest line of text.  
       \return int _maxHeightText*/  
   int dlGetMaxTextHeight() const { return(_maxHeightText); }  
  
   /** Return the largest pixel row number which bounds text.  
       \return int _maxYText*/  
   int dlGetMaxTextRow() const { return(_maxYText); }  
  
   /** Return the pixel width of the widest line of text.  
       \return int _maxWidthText*/  
   int dlGetMaxTextWidth() const { return(_maxWidthText); }  
  
   /** Return the smallest y-ntercept value for the central text lines.  
       \return float _minIntercept*/  
   float dlGetMinIntercept() const { return(_minIntercept); }  
  
   /** Return the smallest slope value for the central text lines.  
       \return float _minSlope*/  
   float dlGetMinSlope() const { return(_minSlope); }  
  
   /** Return the smallest pixel column number which bounds text.  
       \return int _minXText*/  
   int dlGetMinTextCol() const { return(_minXText); }  
  
   /** Return the pixel height of the shortest line of text.  
       \return int _minHeightText*/  
   int dlGetMinTextHeight() const { return(_minHeightText); }  
  
   /** Return the smallest pixel column number which bounds text.  
       \return int _minYText*/  
   int dlGetMinTextRow() const { return(_minYText); }  
  
   /** Return the pixel height of the narrowest line of text.  
       \return int _minWidthText*/  
   int dlGetMinTextWidth() const { return(_minWidthText); }  
  
   /** Return the variance of the slope values for the central text lines.  
       This value is the population variance which is also the second  
       moment of the sample about its mean.  
       \return float _varSlope*/  
   float dlGetSlopeVariance() const { return(_varSlope); }  
  
   /** Return the variance of the pixel text heights.  This value is the  
       population variance which is also the second moment of the sample  
       about its mean.  
       \return float _varHeightText*/  
   float dlGetTextVarianceHeight()  
      const { return(_varHeightText); }  
  
   /** Return the variance of the pixel text widths.  This value is the  
       population variance which is also the second moment of the sample  
       about its mean.  
       \return float _varWidthText*/  
   float dlGetTextVarianceWidth()  
      const { return(_varWidthText); }  
  
   /***********************************************************************  
            SET METHODS  
   ***********************************************************************/  
  
   /** Only save those connected components that have a density less than  
       or equal to dens; the density is the ratio of pixels that are  
       connected components to the total number of connected componets in  
       the bounding box that contains the connected component.  After  
       calling this method, the dlRecalculate() method must be run in order  
       for the new parameter to take effect.  
       \param dens Density upper threshold. */  
   void dlSetFilterCCDensity(float dens)  
      { _densityCCLess = dens; }  
  
   /** Only save those connected components that have a normalized pixel  
       height between low and high inclusive.  After calling this method,  
       the dlRecalculate() method must be run in order for the new  
       parameter to take effect.  
       \param low Lower pixel height threshold.  
       \param high Upper pixel height threshold. */  
   void dlSetFilterCCHeightBetween(int low, int high)  
      { _lowCCHeight = low; _highCCHeight = high; }  
  
   /** Only save those connected components that have a normalized pixel  
       width between low and high inclusive.  After calling this method,  
       the dlRecalculate() method must be run in order for the new  
       parameter to take effect.  
       \param low Lower pixel width threshold.  
       \param high Upper pixel width threshold. */  
   void dlSetFilterCCWidthBetween(int low, int high)  
      { _lowCCWidth = low; _highCCWidth = high; }  
  
   /** Any connected component whose height is more than max times the  
       median height is removed.  After calling this method, the  
       dlRecalculate() method must be run in order for the new parameter to  
       take effect.  
       \param max Upper height factor threshold. */  
   void dlSetMaxToMedianRatio(float max)  
      { _maxToMedianRatio = max; }  
  
   /** Only count text lines whose normalized average (arithmetic mean)  
       pixel heights are greater than min.  After calling this method, the  
       dlRecalculate() method must be run in order for the new parameter to  
       take effect.  
       \param min Lower pixel height threshold. */  
   void dlSetMinAvgHeight(int min)  
      { _minAvgHeight = min; }  
  
   static string dlGetVersionInfo();

   protected:  
   /***********************************************************************  
    GET METHOD VARIABLES  
   ***********************************************************************/  
  
   /** The pixel height of the average (arithmetic mean) line of text. */  
   float _avgHeightText;  
  
   /** The average (arithmetic mean) slope value for the central text  
       lines. */  
   float _avgSlope;  
  
   /** The pixel width of the average (arithmetic mean) line of text. */  
   float _avgWidthText;  
  
   /** The total number of text lines. */  
   int _lineCount;  
  
   /** The pixel height of the tallest line of text. */  
   int _maxHeightText;  
  
   /** The largest y-intercept value for a central text line. */  
   float _maxIntercept;  
  
   /** The largest slope value for a central text line. */  
   float _maxSlope;  
  
   /** The pixel width of the widest line of text. */  
   int _maxWidthText;  
  
   /** The largest (rightmost) column number which bounds text. */  
   int _maxXText;  
  
   /** The largest (downwardmost) row number which bounds text. */  
   int _maxYText;  
  
   /** The pixel height of the shortest line of text. */  
   int _minHeightText;  
  
   /** The smallest y-intercept value for a central text line. */  
   float _minIntercept;  
  
   /** The smallest slope value for a central text line. */  
   float _minSlope;  
  
   /** The pixel width of the shortest line of text. */  
   int _minWidthText;  
  
   /** The smallest (leftmost) column number which bounds text. */  
   int _minXText;  
  
   /** The smallest (uppermost) row number which bounds text. */  
   int _minYText;  
  
   /** The variance of the text heights. */  
   float _varHeightText;  
  
   /** The variance of the slope values for the central text lines. */  
   float _varSlope;  
  
   /** The variance of the text widths. */  
   float _varWidthText;  
  
   /***********************************************************************  
    SET METHOD VARIABLES  
   ***********************************************************************/  
  
   /** Connected components to be used for text line identification must  
       have a ratio of connected components pixels to pixel count in the  
       bounding box area less than or equal to this value. */  
   float _densityCCLess;  
  
   /** Connected components to be used for text line identification must  
       have a height less than or equal to this value. */  
   int _highCCHeight;  
  
   /** Connected components to be used for text line identification must  
       have a width less than or equal to this value. */  
   int _highCCWidth;  
  
   /** Connected components to be used for text line identification must  
       have a height greater than or equal to this value. */  
   int _lowCCHeight;  
  
   /** Connected components to be used for text line identification must  
       have a width greater than or equal to this value. */  
   int _lowCCWidth;  
  
   /** Connected components to be used for text line identification must  
       have a height that is not more than the median height times this  
       value. */  
   float _maxToMedianRatio;  
  
   /** The only text lines considered are those which are greater than this  
       value in average (arithmetic mean) pixel height. */  
   int _minAvgHeight;  
  
   private:  
   /***********************************************************************  
    VARIABLES  
   ***********************************************************************/  
  
   /** The average (arithmetic mean) pixel height of a connected  
       component. */  
   float avgHeightCC;  
  
   /** The average (arithmetic mean) pixel width of a connected  
       component. */  
   float avgWidthCC;  
  
   /** The number of filtered connected components in emuBW. */  
   int emuCCSize;  
  
   /** The input image. */  
   DLImage emuBW;  
  
   /** Information for filtered connected components of emuBW. */  
   DLConnectedComponents emuCC;  
  
   /** A list of the connected components of emuBW. */  
   vector <DLComponent> emuCCList;  
  
   /** A list of the filtered connected components of emuBW. */  
   vector <DLComponent> emuCCListFilt;  

   /** This value is a replacement for
       emuCCListFilt[emuCCListFilt.size() - 1].dlGetRowMax().  In the original
       code, the method dlSetRowMax() was called and later forbidden. */
   int emuCCListFiltRowMaxVal;
  
   /** The output image which is the input image with the text line bounding  
       box information superimposed. */  
   DLImage emuColor;  
  
   /** The number of filtered connected components in emuCC. */  
   int emuSize;  
  
   /** Thresholds should be scaled by this value which is a function of  
       resolution; original code assumed 200dpi.  In the event resolution  
       cannot be read, 200dpi will be assumed. */  
   float factor;  
  
   /***********************************************************************  
    METHODS  
   ***********************************************************************/  
  
   /** Obtain information about a line segment.  
       \param *line_segs Line segment array.  
       \param *textLines Text line array.  
       \param *line_count Text line count. */  
   void dlBuildLineInfo(int *line_segs, DLlineInfo *textLines,  
                        int *line_count);  
  
   /** Use dress-right-dress algorithm to find the horizontal right  
       neighbors of connected components into lines of text.  The average  
       (arithmetic mean) component height and widths for the page are used  
       to prevent neighbors where the vertical separation between connected  
       components exceeds the average connected component height or the  
       horizontal separation is more the 5 times the average connected  
       component width.  
  
       This method is similar to the TwoBit function tbCalcLineOrientation  
       except that only horizontal neighbors are computed and there is a  
       threshold on the vertical separation of neighbors (based on the  
       average connnected component height).  
  
       This version uses a triangle region to the right of CC.  
       <DFN>  
       <br><br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-&nbsp;&nbsp;&nbsp;&nbsp;|<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-&nbsp;&nbsp;_____|&nbsp;SLOPE<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;TO&nbsp;THE&nbsp;RIGHT<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;x_________________________>&nbsp;DD&nbsp;(radius)<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;\\&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;BELOW&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\  
       &nbsp;&nbsp;<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\<br>  
       &nbsp;&nbsp;<br>  
       &nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\\<br>  
       </DFN><br>  
       \param *iList Filtered connected component indices.  
       \param **iEdges Filtered connected component indices formatted  
       differently than iList.  These are the actual edges.  Between  
       edges 0 is a horizontal edge and 1 is a vertical edge, the  
       array is ended by a -2. */  
   // The below diagram is the same as the above but not formatted for  
   // Doxygen.  
   /*  
       -  
       -  
       -  
       -    |  
       -  _____| SLOPE  
       -  
       -  
       -  
       -         TO THE RIGHT  
       x_________________________> DD (radius)  
       -  
       |  \   -  
       -  
       |      \       -  
       -  
       |          \           -  
       -  
       |  BELOW       \               -  
       -  
       |                  \  
  
       |                    \  
  
       |                        \  
   */  
   void dlCalcHorizontalNeighbors(int *iList, int **iEdges);  
  
   /** Combine lines.  
       \param *originalList Collection of line information.  
       \param *newList Empty structure of line information.  
       \param line1 Index of line in orignalList.  
       \param line2 Index of line in originalList.  
       \param newIndex Index of new line. */  
   void dlCombineLines(DLlineInfo *originalList, DLlineInfo *newList,  
                       int line1, int line2, int newIndex);  
  
   /** The main method that calls the others.  It mainly fills in  
       protected values to be returned from "get" methods. */  
   void dlCompLineInfo();  
  
   /** Compute text line information.  
       \param **textLines Collection of line information.  
       \param maxIterations Number of merge iterations.  
       \param *iList Filtered connected component indices.  
       \param *finalLineCount Text line count. */  
   void dlComputeLines(DLlineInfo **textLines, int maxIterations,  
                       int *iList, int *finalLineCount);  
  
   /** Copy values from one DLlineInfo structure to another.  
       \param *originalList Collection of line information.  
       \param *newList Empty structure of line information.  
       \param line1 Index of array in orignalList containing line  
       information.  
       \param newIndex Index of array in newList to which line  
       information should be copied. */  
   void dlCopyToNewLineList(DLlineInfo *originalList, DLlineInfo *newList,  
                            int line1, int newIndex);  
  
   /** Store filtered connected component information.  
       \param **iList Filtered connected component indices. */  
   void dlCreateComponentList(int **iList);  
  
   /** Free allocated memory associated with line information.  
       \param *textLines Collection of line information to free.  
       \param line_count Size of *textLines. */  
   void dlFreeLineInfo(DLlineInfo *textLines, int line_count);  
  
   /** Compute the minimum distance between components.  
       \param *comp1 Connected component information.  
       \param *comp2 Connected component information.  
       \param *min_dist The minimum distance between comp1 and  
       comp2. */  
   void dlGetCompMinDistance(DLComponent *comp1, DLComponent *comp2,  
                             float *min_dist);  
  
   /** Calculate the slope and y-intercept of a line approximated by a  
       collection of connected components.  
       \param *x Horizontal centroids of connected components.  
       \param *y Vertical centroids of connected components.  
       \param size Connected component count.  
       \param *slope The slope of the line approximated by *x and *y.  
       \param *intercept The y-intercept of the line approximated  
       by *x and *y. */  
   void dlLeastSquaresFit(float *x, float *y, int size, float *slope,  
                          float *intercept);  
  
   /** Compute the minimum distance between two line segments.  
       \param a1 Slope of line segment 1.  
       \param b1 Y-intercept of line segment 1.  
       \param a2 Slope of line segment 2.  
       \param b2 Y-intercept of line segment 2.  
       \param l1x1 Minimum horizontal value for line segment 1.  
       \param l1x2 Maximum horizontal value for line segment 1.  
       \param l2x1 Minimum horizontal value for line segment 2.  
       \param l2x2 Maximum horizontal value for line segment 2.  
       \param *lineSepX1 Horizontal value for line segment 1 that  
       yields the greatest vertical distance between line segments.  
       \param *lineSepX2 Horizontal value for line segment 2 that  
       yields the greatest vertical distance between line  
       segments.  
       \return double The minimum distance between two line segments. */  
   double dlLineDistance(double a1, double b1, double a2, double b2,  
                         double l1x1, double l1x2, double l2x1, double l2x2, double *lineSepX1,  
                         double *lineSepX2);  
  
   /** This method will determine the maximum distance from the  
       determined textLine, both below and above it, for each individual  
       component.  It also will place the largest distance, above and below  
       the line, in the DLlineInfo textLines, under the fields  
       upper_ext and lower_ext.  Lastly, it places each  
       individual component's maximum distance, both above and below the  
       line, in separate arrays for future use.  
       \param *textLine Collection of line information.  
       \param line_count Line count for *textLine. */  
   void dlLineExtension(DLlineInfo *textLine, int line_count);  

   /** Generate an output image with the text lines highlighted
       in color
   */

   void dlGenerateAnnotatedImage(DLImage image, char *fname,
                                 DLColorMapValue color);
  
   /** Compute text line statistics.  
       \param *textLine Collection of line information. */  
   void dlLineStats(DLlineInfo *textLine);  
  
   /** Project the endpoints of one line onto another.  
       \param l1x1 Minimum horizontal value for line segment 1.  
       \param l1y1 The y-coordinate that cooresponds with l1x1.  
       \param l1x2 Maximum horizontal value for line segment 1.  
       \param l1y2 The y-coordinate that cooresponds with l1x2.  
       \param l2x1 Minimum horizontal value for line segment 2.  
       \param l2y1 The y-coordinate that cooresponds with l2x1.  
       \param l2x2 Maximum horizontal value for line segment 2.  
       \param l2y2 The y-coordinate that cooresponds with l2x2.  
       \param *limits1 The lengths from the origin to the  
       endpoints of segment 1.  
       \param *projects2 Line projection distances.  
       \param *distance2 Distances between line segment projected  
       endpoints. */  
   void dlProject(double l1x1, double l1y1, double l1x2, double l1y2,  
                  double l2x1, double l2y1, double l2x2, double l2y2, double *limits1,  
                  double *projects2, double *distance2);  
  
   /** Add pairwise connected component information to node list.  
       \param *N1 Collection of node information.  
       \param i1 Index of connected component.  
       \param j1 Index of connected component. */  
   void dlPutOnList(DLnodes *N1, int i1, int j1);  
  
   /** This method is used to remove outlier connected components from  
       line computation.  
       \param *newTextLines Collection of line information.  
       \param line_count Number of lines in *newTextLines. */  
   void dlRemoveOutliers(DLlineInfo *newTextLines, int line_count);  
}; // end getLines class  
  
#endif  
